"""
This example demonstrates how to get an authorization token for a user and then use it to make a request to the Google API on behalf of the user.
"""

from arcadepy import Arcade
from google.oauth2.credentials import Credentials  # pip install google-auth
from googleapiclient.discovery import build  # pip install google-api-python-client


def get_auth_token(client: Arcade, user_id: str) -> str:
    """Get an authorization token for a user.

    In this example, we are
        1. Starting the authorization process for the Gmail Readonly scope
        2. Waiting for the user to authorize the scope
        3. Getting the authorization token
        4. Using the authorization token to make a request to the Google API on behalf of the user
    """
    # Start the authorization process
    auth_response = client.auth.start(
        user_id, "google", scopes=["https://www.googleapis.com/auth/gmail.readonly"]
    )

    if auth_response.status != "completed":
        print(f"Click this link to authorize: {auth_response.url}")
        auth_response = client.auth.wait_for_completion(auth_response)

    return auth_response.context.token


def use_auth_token(token: str) -> None:
    """Use an authorization token to make a request to the Google API on behalf of a user.

    In this example, we are
        1. Using the authorization token that we got from the authorization process to make a request to the Google API
        client.auth.wait_for_completion(auth_response)
    """
    # Use the token from the authorization response
    creds = Credentials(token)
    service = build("gmail", "v1", credentials=creds)

    # Now you can use the Google API
    results = service.users().labels().list(userId="me").execute()
    labels = results.get("labels", [])
    print("Labels:", labels)


if __name__ == "__main__":
    cloud_host = "https://api.arcade.dev"

    client = Arcade(
        base_url=cloud_host,  # Alternatively, use http://localhost:9099 if you are running Arcade locally, or any base_url if you're hosting elsewhere
    )

    user_id = "you@example.com"

    token = get_auth_token(client, user_id)
    use_auth_token(token)



from typing import Annotated, Optional

from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import Dropbox
from arcade_tdk.errors import ToolExecutionError

from arcade_dropbox.constants import Endpoint
from arcade_dropbox.exceptions import DropboxApiError
from arcade_dropbox.utils import parse_dropbox_path, send_dropbox_request


@tool(
    requires_auth=Dropbox(
        scopes=["files.content.read"],
    )
)
async def download_file(
    context: ToolContext,
    file_path: Annotated[
        Optional[str],
        "The path of the file to download. E.g. '/AcmeInc/Reports/Q1_2025.txt'. Defaults to None.",
    ] = None,
    file_id: Annotated[
        Optional[str],
        "The ID of the file to download. E.g. 'id:a4ayc_80_OEAAAAAAAAAYa'. Defaults to None.",
    ] = None,
) -> Annotated[dict, "Contents of the specified file"]:
    """Downloads the specified file.

    Note: either one of `file_path` or `file_id` must be provided.
    """
    if not file_path and not file_id:
        raise ToolExecutionError("Either `file_path` or `file_id` must be provided.")

    if file_path and file_id:
        raise ToolExecutionError("Only one of `file_path` or `file_id` can be provided.")

    try:
        result = await send_dropbox_request(
            context.get_auth_token_or_empty(),
            endpoint=Endpoint.DOWNLOAD_FILE,
            path=parse_dropbox_path(file_path) or file_id,
        )
    except DropboxApiError as api_error:
        return {"error": api_error.message}

    return {"file": result}



import logging
from typing import Any, cast

from google.oauth2.credentials import Credentials
from googleapiclient.discovery import Resource, build

from arcade_google_contacts.constants import DEFAULT_SEARCH_CONTACTS_LIMIT

logging.basicConfig(
    level=logging.DEBUG,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)

logger = logging.getLogger(__name__)


def build_people_service(auth_token: str | None) -> Resource:  # type: ignore[no-any-unimported]
    """
    Build a People service object.
    """
    auth_token = auth_token or ""
    return build("people", "v1", credentials=Credentials(auth_token))


def search_contacts(service: Any, query: str, limit: int | None) -> list[dict[str, Any]]:
    """
    Search the user's contacts in Google Contacts.
    """
    response = (
        service.people()
        .searchContacts(
            query=query,
            pageSize=limit or DEFAULT_SEARCH_CONTACTS_LIMIT,
            readMask=",".join([
                "names",
                "nicknames",
                "emailAddresses",
                "phoneNumbers",
                "addresses",
                "organizations",
                "biographies",
                "urls",
                "userDefined",
            ]),
        )
        .execute()
    )

    return cast(list[dict[str, Any]], response.get("results", []))



from typing import Annotated

from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import Spotify

from arcade_spotify.tools.utils import (
    get_url,
    send_spotify_request,
)


@tool(requires_auth=Spotify())
async def get_track_from_id(
    context: ToolContext,
    track_id: Annotated[str, "The Spotify ID of the track"],
) -> Annotated[dict, "Information about the track"]:
    """Get information about a track"""
    url = get_url("tracks_get_track", track_id=track_id)

    response = await send_spotify_request(context, "GET", url)
    response.raise_for_status()
    return dict(response.json())



from typing import Annotated, Any

from arcade_tdk import ToolContext, tool
from firecrawl import FirecrawlApp

from arcade_web.tools.models import Formats


# TODO: Support actions. This would enable clicking, scrolling, screenshotting, etc.
# TODO: Support extract.
# TODO: Support headers param?
@tool(requires_secrets=["FIRECRAWL_API_KEY"])
async def scrape_url(
    context: ToolContext,
    url: Annotated[str, "URL to scrape"],
    formats: Annotated[
        list[Formats] | None, "Formats to retrieve. Defaults to ['markdown']."
    ] = None,
    only_main_content: Annotated[
        bool | None,
        "Only return the main content of the page excluding headers, navs, footers, etc.",
    ] = True,
    include_tags: Annotated[list[str] | None, "List of tags to include in the output"] = None,
    exclude_tags: Annotated[list[str] | None, "List of tags to exclude from the output"] = None,
    wait_for: Annotated[
        int | None,
        "Specify a delay in milliseconds before fetching the content, allowing the page "
        "sufficient time to load.",
    ] = 10,
    timeout: Annotated[int | None, "Timeout in milliseconds for the request"] = 30000,
) -> Annotated[dict[str, Any], "Scraped data in specified formats"]:
    """Scrape a URL using Firecrawl and return the data in specified formats."""

    api_key = context.get_secret("FIRECRAWL_API_KEY")

    formats = formats or [Formats.MARKDOWN]

    app = FirecrawlApp(api_key=api_key)
    params = {
        "formats": formats,
        "onlyMainContent": only_main_content,
        "includeTags": include_tags or [],
        "excludeTags": exclude_tags or [],
        "waitFor": wait_for,
        "timeout": timeout,
    }
    response = app.scrape_url(url, params=params)

    return dict(response)


# TODO: Support scrapeOptions.
@tool(requires_secrets=["FIRECRAWL_API_KEY"])
async def crawl_website(
    context: ToolContext,
    url: Annotated[str, "URL to crawl"],
    exclude_paths: Annotated[list[str] | None, "URL patterns to exclude from the crawl"] = None,
    include_paths: Annotated[list[str] | None, "URL patterns to include in the crawl"] = None,
    max_depth: Annotated[int, "Maximum depth to crawl relative to the entered URL"] = 2,
    ignore_sitemap: Annotated[bool, "Ignore the website sitemap when crawling"] = True,
    limit: Annotated[int, "Limit the number of pages to crawl"] = 10,
    allow_backward_links: Annotated[
        bool,
        "Enable navigation to previously linked pages and enable crawling "
        "sublinks that are not children of the 'url' input parameter.",
    ] = False,
    allow_external_links: Annotated[bool, "Allow following links to external websites"] = False,
    webhook: Annotated[
        str | None,
        "The URL to send a POST request to when the crawl is started, updated and completed.",
    ] = None,
    async_crawl: Annotated[bool, "Run the crawl asynchronously"] = True,
) -> Annotated[dict[str, Any], "Crawl status and data"]:
    """
    Crawl a website using Firecrawl. If the crawl is asynchronous, then returns the crawl ID.
    If the crawl is synchronous, then returns the crawl data.
    """

    api_key = context.get_secret("FIRECRAWL_API_KEY")

    app = FirecrawlApp(api_key=api_key)
    params = {
        "limit": limit,
        "excludePaths": exclude_paths or [],
        "includePaths": include_paths or [],
        "maxDepth": max_depth,
        "ignoreSitemap": ignore_sitemap,
        "allowBackwardLinks": allow_backward_links,
        "allowExternalLinks": allow_external_links,
    }
    if webhook:
        params["webhook"] = webhook

    if async_crawl:
        response = app.async_crawl_url(url, params=params)
        if (
            "url" in response
        ):  # Url isn't clickable, so removing it since only the ID is needed to check status
            del response["url"]
    else:
        response = app.crawl_url(url, params=params)

    return dict(response)


@tool(requires_secrets=["FIRECRAWL_API_KEY"])
async def get_crawl_status(
    context: ToolContext,
    crawl_id: Annotated[str, "The ID of the crawl job"],
) -> Annotated[dict[str, Any], "Crawl status information"]:
    """
    Get the status of a Firecrawl 'crawl' that is either in progress or recently completed.
    """

    api_key = context.get_secret("FIRECRAWL_API_KEY")

    app = FirecrawlApp(api_key=api_key)
    crawl_status = app.check_crawl_status(crawl_id)

    if "data" in crawl_status:
        del crawl_status["data"]

    return dict(crawl_status)


# TODO: Support responses greater than 10 MB. If the response is greater than 10 MB,
#       then the Firecrawl API response will have a next_url field.
@tool(requires_secrets=["FIRECRAWL_API_KEY"])
async def get_crawl_data(
    context: ToolContext,
    crawl_id: Annotated[str, "The ID of the crawl job"],
) -> Annotated[dict[str, Any], "Crawl data information"]:
    """
    Get the data of a Firecrawl 'crawl' that is either in progress or recently completed.
    """

    api_key = context.get_secret("FIRECRAWL_API_KEY")

    app = FirecrawlApp(api_key=api_key)
    crawl_data = app.check_crawl_status(crawl_id)

    return dict(crawl_data)


@tool(requires_secrets=["FIRECRAWL_API_KEY"])
async def cancel_crawl(
    context: ToolContext,
    crawl_id: Annotated[str, "The ID of the asynchronous crawl job to cancel"],
) -> Annotated[dict[str, Any], "Cancellation status information"]:
    """
    Cancel an asynchronous crawl job that is in progress using the Firecrawl API.
    """

    api_key = context.get_secret("FIRECRAWL_API_KEY")

    app = FirecrawlApp(api_key=api_key)
    cancellation_status = app.cancel_crawl(crawl_id)

    return dict(cancellation_status)


@tool(requires_secrets=["FIRECRAWL_API_KEY"])
async def map_website(
    context: ToolContext,
    url: Annotated[str, "The base URL to start crawling from"],
    search: Annotated[str | None, "Search query to use for mapping"] = None,
    ignore_sitemap: Annotated[bool, "Ignore the website sitemap when crawling"] = True,
    include_subdomains: Annotated[bool, "Include subdomains of the website"] = False,
    limit: Annotated[int, "Maximum number of links to return"] = 5000,
) -> Annotated[dict[str, Any], "Website map data"]:
    """
    Map a website from a single URL to a map of the entire website.
    """

    api_key = context.get_secret("FIRECRAWL_API_KEY")

    app = FirecrawlApp(api_key=api_key)
    params: dict[str, Any] = {
        "ignoreSitemap": ignore_sitemap,
        "includeSubdomains": include_subdomains,
        "limit": limit,
    }
    if search:
        params["search"] = search

    map_result = app.map_url(url, params=params)

    return dict(map_result)



DEFAULT_MESSAGE_FIELDS = [
    "bccRecipients",
    "body",
    "ccRecipients",
    "conversationId",
    "conversationIndex",
    "flag",
    "from",
    "hasAttachments",
    "importance",
    "isDraft",
    "isRead",
    "receivedDateTime",
    "replyTo",
    "subject",
    "toRecipients",
    "webLink",
]



from unittest.mock import patch

import pytest
from arcade_tdk import ToolContext, ToolSecretItem

from arcade_google_jobs.exceptions import LanguageNotFoundError
from arcade_google_jobs.tools.google_jobs import search_jobs


@pytest.fixture
def mock_context():
    return ToolContext(secrets=[ToolSecretItem(key="serp_api_key", value="fake_api_key")])


@pytest.mark.asyncio
@patch("arcade_google_jobs.utils.SerpClient")
async def test_search_jobs_success(mock_serp_client, mock_context):
    mock_serp_client_instance = mock_serp_client.return_value
    mock_serp_client_instance.search().as_dict.return_value = {
        "jobs_results": [
            {"title": "Job 1", "link": "http://example.com/1"},
            {"title": "Job 2", "link": "http://example.com/2"},
        ]
    }

    result = await search_jobs(mock_context, "engenheiro de software", "Brazil", "pt", 10, None)
    assert result == {
        "jobs": [
            {"title": "Job 1", "link": "http://example.com/1"},
            {"title": "Job 2", "link": "http://example.com/2"},
        ],
        "next_page_token": None,
    }


@pytest.mark.asyncio
@patch("arcade_google_jobs.utils.SerpClient")
async def test_search_jobs_success_with_custom_language_and_location(
    mock_serp_client, mock_context
):
    mock_serp_client_instance = mock_serp_client.return_value
    mock_serp_client_instance.search().as_dict.return_value = {
        "jobs_results": [
            {"title": "Job 1", "link": "http://example.com/1"},
            {"title": "Job 2", "link": "http://example.com/2"},
        ]
    }

    result = await search_jobs(
        context=mock_context,
        query="engenheiro de software",
        location="Brazil",
        language="pt",
        limit=10,
        next_page_token=None,
    )

    mock_serp_client_instance.search.assert_called_with({
        "engine": "google_jobs",
        "q": "engenheiro de software",
        "hl": "pt",
        "location": "Brazil",
    })

    assert result == {
        "jobs": [
            {"title": "Job 1", "link": "http://example.com/1"},
            {"title": "Job 2", "link": "http://example.com/2"},
        ],
        "next_page_token": None,
    }


@pytest.mark.asyncio
@patch("arcade_google_jobs.utils.SerpClient")
async def test_search_jobs_language_not_found_error(mock_serp_client, mock_context):
    mock_serp_client_instance = mock_serp_client.return_value
    mock_serp_client_instance.search().as_dict.return_value = {
        "jobs_results": [
            {"title": "Job 1", "link": "http://example.com/1"},
            {"title": "Job 2", "link": "http://example.com/2"},
        ]
    }

    with pytest.raises(LanguageNotFoundError):
        await search_jobs(
            context=mock_context,
            query="backend engineer",
            language="invalid_language",
        )



import json
from typing import Any, Optional

import httpx

from arcade_dropbox.constants import (
    API_BASE_URL,
    API_VERSION,
    ENDPOINT_URL_MAP,
    Endpoint,
    EndpointType,
)
from arcade_dropbox.exceptions import DropboxApiError


def build_dropbox_url(endpoint_type: EndpointType, endpoint_path: str) -> str:
    base_url = API_BASE_URL.format(endpoint_type=endpoint_type.value)
    return f"{base_url}/{API_VERSION}/{endpoint_path.strip('/')}"


def build_dropbox_headers(token: Optional[str]) -> dict[str, str]:
    return {"Authorization": f"Bearer {token}"} if token else {}


def build_dropbox_json(**kwargs: Any) -> dict:
    return {key: value for key, value in kwargs.items() if value is not None}


async def send_dropbox_request(
    authorization_token: Optional[str],
    endpoint: Endpoint,
    **kwargs: Any,
) -> Any:
    endpoint_type, endpoint_path = ENDPOINT_URL_MAP[endpoint]
    url = build_dropbox_url(endpoint_type, endpoint_path)
    headers = build_dropbox_headers(authorization_token)
    json_data = build_dropbox_json(**kwargs)

    if json_data.get("cursor"):
        url += "/continue"
        # If cursor is provided, every other argument must be ignored to avoid API error
        json_data = {"cursor": json_data["cursor"]}

    if endpoint_type == EndpointType.CONTENT:
        headers["Dropbox-API-Arg"] = json.dumps(json_data)
        json_data = {}

    async with httpx.AsyncClient() as client:
        request_args: dict[str, Any] = {"url": url, "headers": headers}

        if json_data:
            request_args["json"] = json_data

        response = await client.post(**request_args)

        try:
            data = response.json()
        except Exception:
            data = {}

        if response.status_code != 200:
            raise DropboxApiError(
                status_code=response.status_code,
                error_summary=data.get("error_summary", response.text),
                user_message=data.get("user_message"),
            )

        if endpoint_type == EndpointType.CONTENT:
            data = json.loads(response.headers["Dropbox-API-Result"])
            data = clean_dropbox_entry(data, default_type="file")
            data["content"] = response.text
            return data

        return response.json()


def clean_dropbox_entry(entry: dict, default_type: Optional[str] = None) -> dict:
    return {
        "type": entry.get(".tag", default_type),
        "id": entry.get("id"),
        "name": entry.get("name"),
        "path": entry.get("path_display"),
        "size_in_bytes": entry.get("size"),
        "modified_datetime": entry.get("server_modified"),
    }


def clean_dropbox_entries(entries: list[dict]) -> list[dict]:
    return [clean_dropbox_entry(entry) for entry in entries]


def parse_dropbox_path(path: Optional[str]) -> Optional[str]:
    if not isinstance(path, str):
        return ""

    if not path:
        return ""

    if path in ["/", "\\"]:
        return ""

    # Normalize windows-style paths to unix-style paths
    path = path.replace("\\", "/")

    # Dropbox expects the path to always start with a slash
    return "/" + path.strip("/")



from typing import Any

from arcadepy.types import ToolDefinition
from pydantic import BaseModel, Field, create_model

# Mapping of Arcade value types to Python types
TYPE_MAPPING = {
    "string": str,
    "number": float,
    "integer": int,
    "boolean": bool,
    "array": list,
    "json": dict,
}


def get_python_type(val_type: str) -> Any:
    """Map Arcade value types to Python types.

    Args:
        val_type: The value type as a string.

    Returns:
        Corresponding Python type.
    """
    _type = TYPE_MAPPING.get(val_type)
    if _type is None:
        raise ValueError(f"Invalid value type: {val_type}")
    return _type


def tool_definition_to_pydantic_model(tool_def: ToolDefinition) -> type[BaseModel]:
    """Convert a ToolDefinition's inputs into a Pydantic BaseModel.

    Args:
        tool_def: The ToolDefinition object to convert.

    Returns:
        A Pydantic BaseModel class representing the tool's input schema.
    """
    try:
        fields: dict[str, Any] = {}
        for param in tool_def.input.parameters or []:
            param_type = get_python_type(param.value_schema.val_type)
            if param_type == list and param.value_schema.inner_val_type:  # noqa: E721
                inner_type: type[Any] = get_python_type(param.value_schema.inner_val_type)
                param_type = list[inner_type]  # type: ignore[valid-type]
            param_description = param.description or "No description provided."
            default = ... if param.required else None
            fields[param.name] = (
                param_type,
                Field(default=default, description=param_description),
            )
        return create_model(f"{tool_def.name}Args", **fields)
    except ValueError as e:
        raise ValueError(f"Error converting {tool_def.name} parameters into pydantic model: {e}")



from arcade_search.tools.google_finance import get_stock_historical_data, get_stock_summary
from arcade_search.tools.google_flights import search_one_way_flights, search_roundtrip_flights
from arcade_search.tools.google_hotels import search_hotels
from arcade_search.tools.google_jobs import search_jobs
from arcade_search.tools.google_maps import (
    get_directions_between_addresses,
    get_directions_between_coordinates,
)
from arcade_search.tools.google_search import search_google

__all__ = [
    "search_google",  # Google Search
    "get_stock_summary",  # Google Finance
    "get_stock_historical_data",  # Google Finance
    "search_one_way_flights",  # Google Flights
    "search_roundtrip_flights",  # Google Flights
    "search_hotels",  # Google Hotels
    "get_directions_between_addresses",  # Google Maps
    "get_directions_between_coordinates",  # Google Maps
    "search_jobs",  # Google Jobs
]



from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_github
from arcade_github.tools.activity import list_stargazers, set_starred

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.9,
    warn_threshold=0.95,
)

catalog = ToolCatalog()
# Register the GitHub tools
catalog.add_module(arcade_github)


@tool_eval()
def github_activity_eval_suite() -> EvalSuite:
    """Evaluation suite for GitHub Activity tools."""
    suite = EvalSuite(
        name="GitHub Activity Tools Evaluation Suite",
        system_message="You are an AI assistant that helps users interact with GitHub repositories using the provided tools.",
        catalog=catalog,
        rubric=rubric,
    )

    # Set Starred
    suite.add_case(
        name="Star a repository",
        user_message="Star the test repository that is owned by ArcadeAI.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=set_starred,
                args={
                    "owner": "ArcadeAI",
                    "name": "test",
                    "starred": True,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="owner", weight=0.3),
            BinaryCritic(critic_field="name", weight=0.3),
            BinaryCritic(critic_field="starred", weight=0.4),
        ],
    )

    suite.add_case(
        name="Unstar a repository",
        user_message="Unstar the ArcadeAI/test repository.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=set_starred,
                args={
                    "owner": "ArcadeAI",
                    "name": "test",
                    "starred": False,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="owner", weight=0.3),
            BinaryCritic(critic_field="name", weight=0.3),
            BinaryCritic(critic_field="starred", weight=0.4),
        ],
    )

    suite.add_case(
        name="List stargazers for a repository",
        user_message="List 42 stargazers for the ArcadeAI/arcade-ai repository.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_stargazers,
                args={
                    "owner": "ArcadeAI",
                    "repo": "arcade-ai",
                    "limit": 42,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="owner", weight=0.3),
            BinaryCritic(critic_field="repo", weight=0.3),
            BinaryCritic(critic_field="limit", weight=0.4),
        ],
    )

    suite.add_case(
        name="List stargazers for a repository",
        user_message="List all of the stargazers for the ArcadeAI/arcade-ai repo",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_stargazers,
                args={
                    "owner": "ArcadeAI",
                    "repo": "arcade-ai",
                    "limit": None,
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="owner", weight=0.3),
            BinaryCritic(critic_field="repo", weight=0.3),
            BinaryCritic(critic_field="limit", weight=0.4),
        ],
    )

    return suite



from dataclasses import asdict, dataclass, field


# ------------------------------------------------------
#  Parent types.
#  See Notion API docs for more information:
#  https://developers.notion.com/reference/parent-object
# ------------------------------------------------------
@dataclass
class Parent:
    type: str

    def to_dict(self) -> dict:
        return asdict(self)


@dataclass
class DatabaseParent(Parent):
    database_id: str
    type: str = field(init=False, default="database_id")


@dataclass
class PageParent(Parent):
    page_id: str
    type: str = field(init=False, default="page_id")


@dataclass
class WorkspaceParent(Parent):
    workspace: bool = True
    type: str = field(init=False, default="workspace")


@dataclass
class BlockParent(Parent):
    block_id: str
    type: str = field(init=False, default="block_id")


def create_parent(parent_data: dict) -> Parent:
    """
    Create a parent object from a dictionary.

    See https://developers.notion.com/reference/parent-object for more information
    about the parent object.

    Args:
        parent_data (dict): The dictionary containing the parent data.

    Returns:
        Parent: The parent object.
    """
    parent_type = parent_data.get("type")
    if parent_type == "database_id":
        return DatabaseParent(database_id=parent_data.get("database_id", ""))
    elif parent_type == "page_id":
        return PageParent(page_id=parent_data.get("page_id", ""))
    elif parent_type == "workspace":
        return WorkspaceParent()
    elif parent_type == "block_id":
        return BlockParent(block_id=parent_data.get("block_id", ""))
    else:
        raise ValueError(f"Unknown parent type: {parent_type}")  # noqa: TRY003


# ------------------------------------------------------
#  Property types.
#  See Notion API docs for more information:
#  https://developers.notion.com/reference/property-object
#  and https://developers.notion.com/reference/page-property-values
# ------------------------------------------------------


@dataclass
class PageWithPageParentProperties:
    """Properties for a page that has a parent that is also a page"""

    title: str

    def to_dict(self) -> dict:
        return {
            "title": {
                "title": [
                    {
                        "type": "text",
                        "text": {
                            "content": self.title,
                        },
                    },
                ],
            },
        }


@dataclass
class PageWithDatabaseParentProperties:
    # TODO: Implement when database parent is supported for `create_page` tool
    pass


@dataclass
class DatabaseProperties:
    # TODO: Implement when create_database tool is implemented
    pass



from typing import Any

from arcade_evals import BinaryCritic


class RelativeTimeBinaryCritic(BinaryCritic):
    def evaluate(self, expected: Any, actual: Any) -> dict[str, float | bool]:
        """
        Evaluates whether the expected and actual relative time strings are equivalent after
        casting.

        Args:
            expected: The expected value.
            actual: The actual value to compare, cast to the type of expected.

        Returns:
            dict: A dictionary containing the match status and score.
        """
        try:
            actual_casted = self.cast_actual(expected, actual)
        except TypeError:
            actual_casted = actual

        expected_parts = tuple(map(int, expected.split(":")))
        actual_parts = tuple(map(int, actual_casted.split(":")))

        if len(expected_parts) != 3 or len(actual_parts) != 3:
            return {"match": False, "score": 0.0}

        exp_days, exp_hours, exp_minutes = expected_parts
        act_days, act_hours, act_minutes = actual_parts

        match = exp_days == act_days and exp_hours == act_hours and exp_minutes == act_minutes
        return {"match": match, "score": self.weight if match else 0.0}



from functools import wraps
from typing import Any, Callable


def clean_asana_response(func: Callable[..., Any]) -> Callable[..., Any]:
    def response_cleaner(data: dict[str, Any]) -> dict[str, Any]:
        if "gid" in data:
            data["id"] = data["gid"]
            del data["gid"]

        for k, v in data.items():
            if isinstance(v, dict):
                data[k] = response_cleaner(v)
            elif isinstance(v, list):
                data[k] = [
                    item if not isinstance(item, dict) else response_cleaner(item) for item in v
                ]

        return data

    @wraps(func)
    async def wrapper(*args: Any, **kwargs: Any) -> Any:
        response = await func(*args, **kwargs)
        return response_cleaner(response)

    return wrapper



from typing import Any, Callable

from arcade_hubspot.constants import GLOBALLY_IGNORED_FIELDS
from arcade_hubspot.enums import HubspotObject


def remove_none_values(data: dict) -> dict:
    cleaned: dict[str, Any] = {}
    for key, value in data.items():
        if value is None or key in GLOBALLY_IGNORED_FIELDS:
            continue
        if isinstance(value, dict):
            cleaned_dict = remove_none_values(value)
            if cleaned_dict:
                cleaned[key] = cleaned_dict
        elif isinstance(value, (list, tuple, set)):
            collection_type = type(value)
            cleaned_list = [remove_none_values(item) for item in value]
            if cleaned_list:
                cleaned[key] = collection_type(cleaned_list)
        else:
            cleaned[key] = value
    return cleaned


def prepare_api_search_response(data: dict, object_type: HubspotObject) -> dict:
    response: dict[str, Any] = {
        object_type.plural: [clean_data(company, object_type) for company in data["results"]],
    }

    after = data.get("paging", {}).get("next", {}).get("after")

    if after:
        response["pagination"] = {
            "are_there_more_results?": True,
            "next_page_token": after,
        }
    else:
        response["pagination"] = {
            "are_there_more_results?": False,
        }

    return response


def rename_dict_keys(data: dict, rename: dict) -> dict:
    for old_key, new_key in rename.items():
        if old_key in data:
            data[new_key] = data[old_key]
            data.pop(old_key, None)
    return data


def global_cleaner(clean_func: Callable[[dict], dict]) -> Callable[[dict], dict]:
    def global_cleaner(data: dict) -> dict:
        cleaned_data: dict[str, Any] = {}
        if "hs_object_id" in data:
            cleaned_data["id"] = data["hs_object_id"]
            del data["hs_object_id"]

        data = rename_dict_keys(
            data,
            {
                "hubspot_owner_id": "owner_id",
                "hs_timestamp": "datetime",
                "hs_body_preview": "body",
            },
        )

        for key, value in data.items():
            if key in GLOBALLY_IGNORED_FIELDS or value is None:
                continue

            if isinstance(value, dict):
                cleaned_data[key] = global_cleaner(value)

            elif isinstance(value, (list, tuple, set)):
                cleaned_items = []
                for item in value:
                    if isinstance(item, dict):
                        cleaned_items.append(global_cleaner(item))
                    else:
                        cleaned_items.append(item)
                cleaned_data[key] = cleaned_items
            else:
                cleaned_data[key] = value
        return cleaned_data

    def wrapper(data: dict) -> dict:
        return remove_none_values(global_cleaner(clean_func(data["properties"])))

    return wrapper


def clean_data(data: dict, object_type: HubspotObject) -> dict:
    _mapping = {
        HubspotObject.CALL: clean_call_data,
        HubspotObject.COMPANY: clean_company_data,
        HubspotObject.CONTACT: clean_contact_data,
        HubspotObject.DEAL: clean_deal_data,
        HubspotObject.EMAIL: clean_email_data,
        HubspotObject.MEETING: clean_meeting_data,
        HubspotObject.NOTE: clean_note_data,
        HubspotObject.TASK: clean_task_data,
    }
    try:
        return _mapping[object_type](data)
    except KeyError:
        raise ValueError(f"Unsupported object type: {object_type}")


@global_cleaner
def clean_company_data(data: dict) -> dict:
    data["object_type"] = HubspotObject.COMPANY.value
    data["website"] = data.get("website", data.get("domain"))
    data.pop("domain", None)

    data["address"] = {
        "street": data.get("address"),
        "city": data.get("city"),
        "state": data.get("state"),
        "zip": data.get("zip"),
        "country": data.get("country"),
    }
    data.pop("address", None)
    data.pop("city", None)
    data.pop("state", None)
    data.pop("zip", None)
    data.pop("country", None)

    data["annual_revenue"] = {
        "value": data.get("annualrevenue"),
        "currency": data.get("hs_annual_revenue_currency_code"),
    }
    data.pop("annualrevenue", None)
    data.pop("hs_annual_revenue_currency_code", None)

    rename = {
        "linkedin_company_page": "linkedin_url",
        "numberofemployees": "employee_count",
        "type": "company_type",
        "annualrevenue": "annual_revenue",
        "lifecyclestage": "lifecycle_stage",
        "hs_lead_status": "lead_status",
    }
    data = rename_dict_keys(data, rename)
    return data


@global_cleaner
def clean_contact_data(data: dict) -> dict:
    data["object_type"] = HubspotObject.CONTACT.value
    rename = {
        "lifecyclestage": "lifecycle_stage",
        "hs_lead_status": "lead_status",
        "email": "email_address",
    }
    data = rename_dict_keys(data, rename)
    return data


@global_cleaner
def clean_deal_data(data: dict) -> dict:
    data["object_type"] = HubspotObject.DEAL.value

    if data.get("closedate") or data.get("hs_closed_amount"):
        data["close"] = {
            "is_closed": data.get("hs_is_closed"),
            "date": data.get("closedate"),
            "amount": data.get("hs_closed_amount"),
        }

        if data.get("hs_is_closed_won") in ["true", True]:
            data["close"]["status"] = "won"
            data["close"]["status_reason"] = data.get("closed_won_reason")
        elif data.get("hs_is_closed_lost") in ["true", True]:
            data["close"]["status"] = "lost"
            data["close"]["status_reason"] = data.get("closed_lost_reason")

    if data.get("amount"):
        data["amount"] = {
            "value": data["amount"],
            "currency": data.get("deal_currency_code"),
        }

        if data.get("hs_forecast_probability"):
            data["amount"]["forecast"] = {
                "probability": data["hs_forecast_probability"],
                "expected_value": data.get("hs_forecast_amount"),
            }

    rename = {
        "dealname": "name",
        "dealstage": "stage",
        "dealscore": "score",
        "dealtype": "type",
    }
    data = rename_dict_keys(data, rename)

    data.pop("hs_is_closed", None)
    data.pop("closedate", None)
    data.pop("hs_closed_amount", None)
    data.pop("deal_currency_code", None)
    data.pop("close_won_reason", None)
    data.pop("close_lost_reason", None)
    data.pop("hs_is_closed_won", None)
    data.pop("hs_is_closed_lost", None)
    data.pop("hs_forecast_probability", None)
    data.pop("hs_forecast_amount", None)

    return data


@global_cleaner
def clean_email_data(data: dict) -> dict:
    data["object_type"] = HubspotObject.EMAIL.value
    rename = {
        "hs_body_preview": "body",
        "hs_email_from_raw": "from",
        "hs_email_to_raw": "to",
        "hs_email_bcc_raw": "bcc",
        "hs_email_cc_raw": "cc",
        "hs_email_subject": "subject",
        "hs_email_status": "status",
        "hs_email_associated_contact_id": "contact_id",
    }
    data = rename_dict_keys(data, rename)
    return data


@global_cleaner
def clean_call_data(data: dict) -> dict:
    data["object_type"] = HubspotObject.CALL.value
    rename = {
        "hs_call_direction": "direction",
        "hs_call_status": "status",
        "hs_call_summary": "summary",
        "hs_call_title": "title",
        "hs_call_disposition": "outcome",
    }
    data = rename_dict_keys(data, rename)
    return data


@global_cleaner
def clean_note_data(data: dict) -> dict:
    data["object_type"] = HubspotObject.NOTE.value
    return data


@global_cleaner
def clean_meeting_data(data: dict) -> dict:
    data["object_type"] = HubspotObject.MEETING.value
    rename = {
        "hs_meeting_outcome": "outcome",
        "hs_meeting_location": "location",
        "hs_meeting_start_time": "start_time",
        "hs_meeting_end_time": "end_time",
    }
    data = rename_dict_keys(data, rename)

    data["content"] = {
        "title": data.get("hs_meeting_title"),
        "body": data.get("hs_body_preview"),
    }

    data.pop("hs_meeting_title", None)
    data.pop("hs_body_preview", None)

    data["schedule"] = {
        "start_time": data.get("start_time"),
        "end_time": data.get("end_time"),
    }

    data.pop("start_time", None)
    data.pop("end_time", None)

    return data


@global_cleaner
def clean_task_data(data: dict) -> dict:
    data["object_type"] = HubspotObject.TASK.value

    if data.get("hs_task_priority") == "NONE":
        data["hs_task_priority"] = None

    rename = {
        "hs_task_status": "status",
        "hs_task_priority": "priority",
        "hs_task_missed_due_date": "missed_due_date",
        "hs_task_type": "task_type",
        "hs_associated_contact_labels": "associated_contact",
    }
    data = rename_dict_keys(data, rename)

    data["content"] = {
        "body": data.get("hs_body_preview"),
        "subject": data.get("hs_task_subject"),
    }

    data.pop("hs_body_preview", None)
    data.pop("hs_task_subject", None)

    data["schedule"] = {
        "datetime": data.get("hs_timestamp"),
        "is_overdue": data.get("hs_task_is_overdue"),
    }

    data.pop("hs_timestamp", None)
    data.pop("hs_task_is_overdue", None)

    return data



import base64
import json

from arcade_tdk import ToolContext, ToolMetadataKey
from arcade_tdk.errors import ToolExecutionError


def generate_google_file_picker_url(context: ToolContext) -> dict:
    """Generate a Google File Picker URL for user-driven file selection and authorization.

    Generates a URL that directs the end-user to a Google File Picker interface where
    where they can select or upload Google Drive files. Users can grant permission to access their
    Drive files, providing a secure and authorized way to interact with their files.

    This is particularly useful when prior tools (e.g., those accessing or modifying
    Google Docs, Google Sheets, etc.) encountered failures due to file non-existence
    (Requested entity was not found) or permission errors. Once the user completes the file
    picker flow, the prior tool can be retried.

    Returns:
        A dictionary containing the URL and instructions for the llm to instruct the user.
    """
    client_id = context.get_metadata(ToolMetadataKey.CLIENT_ID)
    client_id_parts = client_id.split("-")
    if not client_id_parts:
        raise ToolExecutionError(
            message="Invalid Google Client ID",
            developer_message=f"Google Client ID '{client_id}' is not valid",
        )
    app_id = client_id_parts[0]
    cloud_coordinator_url = context.get_metadata(ToolMetadataKey.COORDINATOR_URL).strip("/")

    config = {
        "auth": {
            "client_id": client_id,
            "app_id": app_id,
        },
    }
    config_json = json.dumps(config)
    config_base64 = base64.urlsafe_b64encode(config_json.encode("utf-8")).decode("utf-8")
    url = f"{cloud_coordinator_url}/google/drive_picker?config={config_base64}"

    return {
        "url": url,
        "llm_instructions": (
            "Instruct the user to click the following link to open the Google Drive File Picker. "
            f"This will allow them to select files and grant access permissions: {url}"
        ),
    }



from enum import Enum


class GoogleMapsTravelMode(Enum):
    BEST = "best"
    DRIVING = "driving"
    MOTORCYCLE = "motorcycle"
    PUBLIC_TRANSPORTATION = "public_transportation"
    WALKING = "walking"
    BICYCLE = "bicycle"
    FLIGHT = "flight"

    def to_api_value(self) -> int:
        _map = {
            str(self.BEST): 6,
            str(self.DRIVING): 0,
            str(self.MOTORCYCLE): 9,
            str(self.PUBLIC_TRANSPORTATION): 3,
            str(self.WALKING): 2,
            str(self.BICYCLE): 1,
            str(self.FLIGHT): 4,
        }
        return _map[str(self)]


class GoogleMapsDistanceUnit(Enum):
    KM = "km"
    MILES = "mi"

    def to_api_value(self) -> int:
        _map = {
            str(self.KM): 0,
            str(self.MILES): 1,
        }
        return _map[str(self)]



from typing import Annotated, Any

from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import Asana
from arcade_tdk.errors import ToolExecutionError

from arcade_asana.constants import TAG_OPT_FIELDS, TagColor
from arcade_asana.models import AsanaClient
from arcade_asana.utils import (
    get_next_page,
    get_unique_workspace_id_or_raise_error,
    remove_none_values,
)


@tool(requires_auth=Asana(scopes=["default"]))
async def get_tag_by_id(
    context: ToolContext,
    tag_id: Annotated[str, "The ID of the Asana tag to get"],
) -> Annotated[dict[str, Any], "Get an Asana tag by its ID"]:
    """Get an Asana tag by its ID"""
    client = AsanaClient(context.get_auth_token_or_empty())
    response = await client.get(f"/tags/{tag_id}")
    return {"tag": response["data"]}


@tool(requires_auth=Asana(scopes=["default"]))
async def create_tag(
    context: ToolContext,
    name: Annotated[str, "The name of the tag to create. Length must be between 1 and 100."],
    description: Annotated[
        str | None, "The description of the tag to create. Defaults to None (no description)."
    ] = None,
    color: Annotated[
        TagColor | None, "The color of the tag to create. Defaults to None (no color)."
    ] = None,
    workspace_id: Annotated[
        str | None,
        "The ID of the workspace to create the tag in. If not provided, it will associated the tag "
        "to a current workspace, if there's only one. Otherwise, it will raise an error.",
    ] = None,
) -> Annotated[dict[str, Any], "The created tag."]:
    """Create a tag in Asana"""
    if not 1 <= len(name) <= 100:
        raise ToolExecutionError("Tag name must be between 1 and 100 characters long.")

    workspace_id = workspace_id or await get_unique_workspace_id_or_raise_error(context)

    data = remove_none_values({
        "name": name,
        "notes": description,
        "color": color.value if color else None,
        "workspace": workspace_id,
    })

    client = AsanaClient(context.get_auth_token_or_empty())
    response = await client.post("/tags", json_data={"data": data})
    return {"tag": response["data"]}


@tool(requires_auth=Asana(scopes=["default"]))
async def list_tags(
    context: ToolContext,
    workspace_id: Annotated[
        str | None,
        "The workspace ID to retrieve tags from. Defaults to None. If not provided and the user "
        "has only one workspace, it will use that workspace. If not provided and the user has "
        "multiple workspaces, it will raise an error listing the available workspaces.",
    ] = None,
    limit: Annotated[
        int, "The maximum number of tags to return. Min is 1, max is 100. Defaults to 100."
    ] = 100,
    next_page_token: Annotated[
        str | None,
        "The token to retrieve the next page of tags. Defaults to None (start from the first page "
        "of tags)",
    ] = None,
) -> Annotated[
    dict[str, Any],
    "List tags in an Asana workspace",
]:
    """List tags in an Asana workspace"""
    limit = max(1, min(100, limit))

    workspace_id = workspace_id or await get_unique_workspace_id_or_raise_error(context)

    client = AsanaClient(context.get_auth_token_or_empty())
    response = await client.get(
        "/tags",
        params=remove_none_values({
            "limit": limit,
            "offset": next_page_token,
            "workspace": workspace_id,
            "opt_fields": TAG_OPT_FIELDS,
        }),
    )

    return {
        "tags": response["data"],
        "count": len(response["data"]),
        "next_page": get_next_page(response),
    }



COUNTRY_CODES = {
    "af": "Afghanistan",
    "al": "Albania",
    "dz": "Algeria",
    "as": "American Samoa",
    "ad": "Andorra",
    "ao": "Angola",
    "ai": "Anguilla",
    "aq": "Antarctica",
    "ag": "Antigua and Barbuda",
    "ar": "Argentina",
    "am": "Armenia",
    "aw": "Aruba",
    "au": "Australia",
    "at": "Austria",
    "az": "Azerbaijan",
    "bs": "Bahamas",
    "bh": "Bahrain",
    "bd": "Bangladesh",
    "bb": "Barbados",
    "by": "Belarus",
    "be": "Belgium",
    "bz": "Belize",
    "bj": "Benin",
    "bm": "Bermuda",
    "bt": "Bhutan",
    "bo": "Bolivia",
    "ba": "Bosnia and Herzegovina",
    "bw": "Botswana",
    "bv": "Bouvet Island",
    "br": "Brazil",
    "io": "British Indian Ocean Territory",
    "bn": "Brunei Darussalam",
    "bg": "Bulgaria",
    "bf": "Burkina Faso",
    "bi": "Burundi",
    "kh": "Cambodia",
    "cm": "Cameroon",
    "ca": "Canada",
    "cv": "Cape Verde",
    "ky": "Cayman Islands",
    "cf": "Central African Republic",
    "td": "Chad",
    "cl": "Chile",
    "cn": "China",
    "cx": "Christmas Island",
    "cc": "Cocos (Keeling) Islands",
    "co": "Colombia",
    "km": "Comoros",
    "cg": "Congo",
    "cd": "Congo, the Democratic Republic of the",
    "ck": "Cook Islands",
    "cr": "Costa Rica",
    "ci": "Cote D'ivoire",
    "hr": "Croatia",
    "cu": "Cuba",
    "cy": "Cyprus",
    "cz": "Czech Republic",
    "dk": "Denmark",
    "dj": "Djibouti",
    "dm": "Dominica",
    "do": "Dominican Republic",
    "ec": "Ecuador",
    "eg": "Egypt",
    "sv": "El Salvador",
    "gq": "Equatorial Guinea",
    "er": "Eritrea",
    "ee": "Estonia",
    "et": "Ethiopia",
    "fk": "Falkland Islands (Malvinas)",
    "fo": "Faroe Islands",
    "fj": "Fiji",
    "fi": "Finland",
    "fr": "France",
    "gf": "French Guiana",
    "pf": "French Polynesia",
    "tf": "French Southern Territories",
    "ga": "Gabon",
    "gm": "Gambia",
    "ge": "Georgia",
    "de": "Germany",
    "gh": "Ghana",
    "gi": "Gibraltar",
    "gr": "Greece",
    "gl": "Greenland",
    "gd": "Grenada",
    "gp": "Guadeloupe",
    "gu": "Guam",
    "gt": "Guatemala",
    "gg": "Guernsey",
    "gn": "Guinea",
    "gw": "Guinea-Bissau",
    "gy": "Guyana",
    "ht": "Haiti",
    "hm": "Heard Island and Mcdonald Islands",
    "va": "Holy See (Vatican City State)",
    "hn": "Honduras",
    "hk": "Hong Kong",
    "hu": "Hungary",
    "is": "Iceland",
    "in": "India",
    "id": "Indonesia",
    "ir": "Iran, Islamic Republic of",
    "iq": "Iraq",
    "ie": "Ireland",
    "im": "Isle of Man",
    "il": "Israel",
    "it": "Italy",
    "je": "Jersey",
    "jm": "Jamaica",
    "jp": "Japan",
    "jo": "Jordan",
    "kz": "Kazakhstan",
    "ke": "Kenya",
    "ki": "Kiribati",
    "kp": "Korea, Democratic People's Republic of",
    "kr": "Korea, Republic of",
    "kw": "Kuwait",
    "kg": "Kyrgyzstan",
    "la": "Lao People's Democratic Republic",
    "lv": "Latvia",
    "lb": "Lebanon",
    "ls": "Lesotho",
    "lr": "Liberia",
    "ly": "Libyan Arab Jamahiriya",
    "li": "Liechtenstein",
    "lt": "Lithuania",
    "lu": "Luxembourg",
    "mo": "Macao",
    "mk": "Macedonia, the Former Yugosalv Republic of",
    "mg": "Madagascar",
    "mw": "Malawi",
    "my": "Malaysia",
    "mv": "Maldives",
    "ml": "Mali",
    "mt": "Malta",
    "mh": "Marshall Islands",
    "mq": "Martinique",
    "mr": "Mauritania",
    "mu": "Mauritius",
    "yt": "Mayotte",
    "mx": "Mexico",
    "fm": "Micronesia, Federated States of",
    "md": "Moldova, Republic of",
    "mc": "Monaco",
    "mn": "Mongolia",
    "me": "Montenegro",
    "ms": "Montserrat",
    "ma": "Morocco",
    "mz": "Mozambique",
    "mm": "Myanmar",
    "na": "Namibia",
    "nr": "Nauru",
    "np": "Nepal",
    "nl": "Netherlands",
    "an": "Netherlands Antilles",
    "nc": "New Caledonia",
    "nz": "New Zealand",
    "ni": "Nicaragua",
    "ne": "Niger",
    "ng": "Nigeria",
    "nu": "Niue",
    "nf": "Norfolk Island",
    "mp": "Northern Mariana Islands",
    "no": "Norway",
    "om": "Oman",
    "pk": "Pakistan",
    "pw": "Palau",
    "ps": "Palestinian Territory, Occupied",
    "pa": "Panama",
    "pg": "Papua New Guinea",
    "py": "Paraguay",
    "pe": "Peru",
    "ph": "Philippines",
    "pn": "Pitcairn",
    "pl": "Poland",
    "pt": "Portugal",
    "pr": "Puerto Rico",
    "qa": "Qatar",
    "re": "Reunion",
    "ro": "Romania",
    "ru": "Russian Federation",
    "rw": "Rwanda",
    "sh": "Saint Helena",
    "kn": "Saint Kitts and Nevis",
    "lc": "Saint Lucia",
    "pm": "Saint Pierre and Miquelon",
    "vc": "Saint Vincent and the Grenadines",
    "ws": "Samoa",
    "sm": "San Marino",
    "st": "Sao Tome and Principe",
    "sa": "Saudi Arabia",
    "sn": "Senegal",
    "rs": "Serbia",
    "sc": "Seychelles",
    "sl": "Sierra Leone",
    "sg": "Singapore",
    "sk": "Slovakia",
    "si": "Slovenia",
    "sb": "Solomon Islands",
    "so": "Somalia",
    "za": "South Africa",
    "gs": "South Georgia and the South Sandwich Islands",
    "es": "Spain",
    "lk": "Sri Lanka",
    "sd": "Sudan",
    "sr": "Suriname",
    "sj": "Svalbard and Jan Mayen",
    "sz": "Swaziland",
    "se": "Sweden",
    "ch": "Switzerland",
    "sy": "Syrian Arab Republic",
    "tw": "Taiwan, Province of China",
    "tj": "Tajikistan",
    "tz": "Tanzania, United Republic of",
    "th": "Thailand",
    "tl": "Timor-Leste",
    "tg": "Togo",
    "tk": "Tokelau",
    "to": "Tonga",
    "tt": "Trinidad and Tobago",
    "tn": "Tunisia",
    "tr": "Turkiye",
    "tm": "Turkmenistan",
    "tc": "Turks and Caicos Islands",
    "tv": "Tuvalu",
    "ug": "Uganda",
    "ua": "Ukraine",
    "ae": "United Arab Emirates",
    "uk": "United Kingdom",
    "gb": "United Kingdom",
    "us": "United States",
    "um": "United States Minor Outlying Islands",
    "uy": "Uruguay",
    "uz": "Uzbekistan",
    "vu": "Vanuatu",
    "ve": "Venezuela",
    "vn": "Viet Nam",
    "vg": "Virgin Islands, British",
    "vi": "Virgin Islands, U.S.",
    "wf": "Wallis and Futuna",
    "eh": "Western Sahara",
    "ye": "Yemen",
    "zm": "Zambia",
    "zw": "Zimbabwe",
}


LANGUAGE_CODES = {
    "ar": "Arabic",
    "bn": "Bengali",
    "da": "Danish",
    "de": "German",
    "el": "Greek",
    "en": "English",
    "es": "Spanish",
    "fi": "Finnish",
    "fr": "French",
    "hi": "Hindi",
    "hu": "Hungarian",
    "id": "Indonesian",
    "it": "Italian",
    "ja": "Japanese",
    "ko": "Korean",
    "nl": "Dutch",
    "ms": "Malay",
    "no": "Norwegian",
    "pcm": "Nigerian Pidgin",
    "pl": "Polish",
    "pt": "Portuguese",
    "pt-br": "Portuguese (Brazil)",
    "pt-pt": "Portuguese (Portugal)",
    "ru": "Russian",
    "sv": "Swedish",
    "tl": "Filipino",
    "tr": "Turkish",
    "uk": "Ukrainian",
    "zh": "Chinese",
    "zh-cn": "Chinese (Simplified)",
    "zh-tw": "Chinese (Traditional)",
}


GOOGLE_DOMAIN_BY_COUNTRY_CODE = {
    "ad": "google.ad",
    "ae": "google.ae",
    "al": "google.al",
    "am": "google.am",
    "as": "google.as",
    "at": "google.at",
    "az": "google.az",
    "ba": "google.ba",
    "be": "google.be",
    "bf": "google.bf",
    "bg": "google.bg",
    "bi": "google.bi",
    "bj": "google.bj",
    "bs": "google.bs",
    "bt": "google.bt",
    "by": "google.by",
    "ca": "google.ca",
    "cg": "google.cg",
    "cf": "google.cf",
    "ch": "google.ch",
    "ci": "google.ci",
    "cl": "google.cl",
    "cm": "google.cm",
    "ao": "google.co.ao",
    "bw": "google.co.bw",
    "ck": "google.co.ck",
    "cr": "google.co.cr",
    "id": "google.co.id",
    "il": "google.co.il",
    "in": "google.co.in",
    "jp": "google.co.jp",
    "ke": "google.co.ke",
    "kr": "google.co.kr",
    "ls": "google.co.ls",
    "ma": "google.co.ma",
    "mz": "google.co.mz",
    "nz": "google.co.nz",
    "th": "google.co.th",
    "tz": "google.co.tz",
    "ug": "google.co.ug",
    "uk": "google.co.uk",
    "uz": "google.co.uz",
    "ve": "google.co.ve",
    "vi": "google.co.vi",
    "za": "google.co.za",
    "zm": "google.co.zm",
    "zw": "google.co.zw",
    "us": "google.com",
    "af": "google.com.af",
    "ag": "google.com.ag",
    "ai": "google.com.ai",
    "ar": "google.com.ar",
    "au": "google.com.au",
    "bd": "google.com.bd",
    "bh": "google.com.bh",
    "bn": "google.com.bn",
    "bo": "google.com.bo",
    "br": "google.com.br",
    "bz": "google.com.bz",
    "co": "google.com.co",
    "cu": "google.com.cu",
    "cy": "google.com.cy",
    "do": "google.com.do",
    "ec": "google.com.ec",
    "eg": "google.com.eg",
    "et": "google.com.et",
    "fj": "google.com.fj",
    "gh": "google.com.gh",
    "gi": "google.com.gi",
    "gt": "google.com.gt",
    "hk": "google.com.hk",
    "jm": "google.com.jm",
    "kh": "google.com.kh",
    "kw": "google.com.kw",
    "lb": "google.com.lb",
    "ly": "google.com.ly",
    "mm": "google.com.mm",
    "mt": "google.com.mt",
    "mx": "google.com.mx",
    "my": "google.com.my",
    "na": "google.com.na",
    "ng": "google.com.ng",
    "ni": "google.com.ni",
    "np": "google.com.np",
    "om": "google.com.om",
    "pa": "google.com.pa",
    "pe": "google.com.pe",
    "pg": "google.com.pg",
    "ph": "google.com.ph",
    "pk": "google.com.pk",
    "pr": "google.com.pr",
    "py": "google.com.py",
    "qa": "google.com.qa",
    "sa": "google.com.sa",
    "sb": "google.com.sb",
    "sg": "google.com.sg",
    "sl": "google.com.sl",
    "sv": "google.com.sv",
    "tj": "google.com.tj",
    "tr": "google.com.tr",
    "tw": "google.com.tw",
    "ua": "google.com.ua",
    "uy": "google.com.uy",
    "vc": "google.com.vc",
    "vn": "google.com.vn",
    "cv": "google.cv",
    "cz": "google.cz",
    "de": "google.de",
    "dj": "google.dj",
    "dk": "google.dk",
    "dm": "google.dm",
    "dz": "google.dz",
    "ee": "google.ee",
    "es": "google.es",
    "fi": "google.fi",
    "fm": "google.fm",
    "fr": "google.fr",
    "ga": "google.ga",
    "ge": "google.ge",
    "gl": "google.gl",
    "gm": "google.gm",
    "gp": "google.gp",
    "gr": "google.gr",
    "gy": "google.gy",
    "hn": "google.hn",
    "hr": "google.hr",
    "ht": "google.ht",
    "hu": "google.hu",
    "ie": "google.ie",
    "iq": "google.iq",
    "is": "google.is",
    "it": "google.it",
    "je": "google.je",
    "jo": "google.jo",
    "kg": "google.kg",
    "ki": "google.ki",
    "kz": "google.kz",
    "la": "google.la",
    "li": "google.li",
    "lk": "google.lk",
    "lt": "google.lt",
    "lu": "google.lu",
    "lv": "google.lv",
    "md": "google.md",
    "mg": "google.mg",
    "mk": "google.mk",
    "ml": "google.ml",
    "mn": "google.mn",
    "ms": "google.ms",
    "mu": "google.mu",
    "mv": "google.mv",
    "mw": "google.mw",
    "ne": "google.ne",
    "nl": "google.nl",
    "no": "google.no",
    "nr": "google.nr",
    "nu": "google.nu",
    "pl": "google.pl",
    "ps": "google.ps",
    "pt": "google.pt",
    "ro": "google.ro",
    "rs": "google.rs",
    "ru": "google.ru",
    "rw": "google.rw",
    "sc": "google.sc",
    "se": "google.se",
    "sh": "google.sh",
    "si": "google.si",
    "sk": "google.sk",
    "sm": "google.sm",
    "sn": "google.sn",
    "so": "google.so",
    "sr": "google.sr",
    "td": "google.td",
    "tg": "google.tg",
    "tk": "google.tk",
    "tl": "google.tl",
    "tm": "google.tm",
    "tn": "google.tn",
    "to": "google.to",
    "tt": "google.tt",
    "vg": "google.vg",
    "vu": "google.vu",
    "ws": "google.ws",
}



from typing import Annotated, cast

from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import Slack
from arcade_tdk.errors import ToolExecutionError
from slack_sdk.errors import SlackApiError
from slack_sdk.web.async_client import AsyncWebClient

from arcade_slack.constants import MAX_PAGINATION_SIZE_LIMIT
from arcade_slack.conversation_retrieval import (
    get_channel_by_name,
    get_conversation_by_id,
)
from arcade_slack.message_retrieval import retrieve_messages_in_conversation
from arcade_slack.models import (
    ConversationType,
)
from arcade_slack.user_retrieval import (
    get_users_by_id,
    get_users_by_id_username_or_email,
)
from arcade_slack.utils import (
    async_paginate,
    extract_conversation_metadata,
    populate_users_in_messages,
    raise_for_users_not_found,
)


@tool(
    requires_auth=Slack(
        scopes=[
            "channels:read",
            "groups:read",
            "mpim:read",
            "im:read",
            "users:read",
            "users:read.email",
            "chat:write",
            # "mpim:write",
            "im:write",
        ],
    )
)
async def send_message(
    context: ToolContext,
    message: Annotated[str, "The content of the message to send."],
    channel_name: Annotated[
        str | None,
        "The channel name to send the message to. Prefer providing a conversation_id, "
        "when available, since the performance is better.",
    ] = None,
    conversation_id: Annotated[str | None, "The conversation ID to send the message to."] = None,
    user_ids: Annotated[list[str] | None, "The Slack user IDs of the people to message."] = None,
    emails: Annotated[list[str] | None, "The emails of the people to message."] = None,
    usernames: Annotated[
        list[str] | None,
        "The Slack usernames of the people to message. Prefer providing user_ids and/or emails, "
        "when available, since the performance is better.",
    ] = None,
) -> Annotated[dict, "The response from the Slack API"]:
    """Send a message to a Channel, Direct Message (IM/DM), or Multi-Person (MPIM) conversation

    Provide exactly one of:
    - channel_name; or
    - conversation_id; or
    - any combination of user_ids, usernames, and/or emails.

    In case multiple user_ids, usernames, and/or emails are provided, the tool will open a
    multi-person conversation with the specified people and send the message to it.
    """
    if conversation_id and any([channel_name, user_ids, usernames, emails]):
        raise ToolExecutionError(
            "Provide exactly one of: channel_name, OR conversation_id, OR any combination of "
            "user_ids, usernames, and/or emails."
        )

    if not conversation_id:
        conversation = await get_conversation_metadata(
            context=context,
            channel_name=channel_name,
            user_ids=user_ids,
            usernames=usernames,
            emails=emails,
        )
        conversation_id = conversation["id"]

    slack_client = AsyncWebClient(token=context.get_auth_token_or_empty())
    response = await slack_client.chat_postMessage(channel=cast(str, conversation_id), text=message)
    return {"success": True, "data": response.data}


@tool(
    requires_auth=Slack(
        scopes=[
            "channels:read",
            "groups:read",
            "im:read",
            "mpim:read",
            "users:read",
            "users:read.email",
        ],
    )
)
async def get_users_in_conversation(
    context: ToolContext,
    conversation_id: Annotated[str | None, "The ID of the conversation to get users in."] = None,
    channel_name: Annotated[
        str | None,
        "The name of the channel to get users in. Prefer providing a conversation_id, "
        "when available, since the performance is better.",
    ] = None,
    # The user object is relatively small, so we allow a higher limit.
    limit: Annotated[
        int, "The maximum number of users to return. Defaults to 200. Maximum is 500."
    ] = 200,
    next_cursor: Annotated[str | None, "The cursor to use for pagination."] = None,
) -> Annotated[dict, "Information about each user in the conversation"]:
    """Get the users in a Slack conversation (Channel, DM/IM, or MPIM) by its ID or by channel name.

    Provide exactly one of conversation_id or channel_name. Prefer providing a conversation_id,
    when available, since the performance is better.
    """
    limit = max(1, min(limit, 500))

    if sum({bool(conversation_id), bool(channel_name)}) != 1:
        raise ToolExecutionError("Provide exactly one of conversation_id OR channel_name.")

    auth_token = context.get_auth_token_or_empty()

    if not conversation_id:
        channel = await get_channel_by_name(auth_token, cast(str, channel_name))
        conversation_id = channel["id"]

    slack_client = AsyncWebClient(token=auth_token)
    user_ids, next_cursor = await async_paginate(
        func=slack_client.conversations_members,
        response_key="members",
        limit=limit,
        next_cursor=next_cursor,
        channel=conversation_id,
    )

    response = await get_users_by_id(auth_token, user_ids)

    await raise_for_users_not_found(context, [response])

    return {
        "users": [user for user in response["users"] if not user.get("is_bot")],
        "next_cursor": next_cursor,
    }


@tool(
    requires_auth=Slack(
        scopes=[
            "channels:read",
            "groups:read",
            "mpim:read",
            "im:read",
            "users:read",
            "users:read.email",
            "channels:history",
            "groups:history",
            "mpim:history",
            "im:history",
        ]
    )
)
async def get_messages(
    context: ToolContext,
    conversation_id: Annotated[
        str | None,
        "The ID of the conversation to get messages from. Provide exactly one of conversation_id "
        "OR any combination of user_ids, usernames, and/or emails.",
    ] = None,
    channel_name: Annotated[
        str | None,
        "The name of the channel to get messages from. Prefer providing a conversation_id, "
        "when available, since the performance is better.",
    ] = None,
    user_ids: Annotated[
        list[str] | None, "The IDs of the users in the conversation to get messages from."
    ] = None,
    usernames: Annotated[
        list[str] | None,
        "The usernames of the users in the conversation to get messages from. Prefer providing"
        "user_ids and/or emails, when available, since the performance is better.",
    ] = None,
    emails: Annotated[
        list[str] | None,
        "The emails of the users in the conversation to get messages from.",
    ] = None,
    oldest_relative: Annotated[
        str | None,
        (
            "The oldest message to include in the results, specified as a time offset from the "
            "current time in the format 'DD:HH:MM'"
        ),
    ] = None,
    latest_relative: Annotated[
        str | None,
        (
            "The latest message to include in the results, specified as a time offset from the "
            "current time in the format 'DD:HH:MM'"
        ),
    ] = None,
    oldest_datetime: Annotated[
        str | None,
        (
            "The oldest message to include in the results, specified as a datetime object in the "
            "format 'YYYY-MM-DD HH:MM:SS'"
        ),
    ] = None,
    latest_datetime: Annotated[
        str | None,
        (
            "The latest message to include in the results, specified as a datetime object in the "
            "format 'YYYY-MM-DD HH:MM:SS'"
        ),
    ] = None,
    limit: Annotated[
        # The message object can be relatively large, so we limit maximum to 100
        # to preserve LLM's context window and reduce the likelihood of hallucinations.
        int, "The maximum number of messages to return. Defaults to 20. Maximum is 100."
    ] = 20,
    next_cursor: Annotated[str | None, "The cursor to use for pagination."] = None,
) -> Annotated[
    dict,
    "The messages in a Slack Channel, DM (direct message) or MPIM (multi-person) conversation.",
]:
    """Get messages in a Slack Channel, DM (direct message) or MPIM (multi-person) conversation.

    Provide exactly one of:
    - conversation_id; or
    - channel_name; or
    - any combination of user_ids, usernames, and/or emails.

    To filter messages by an absolute datetime, use 'oldest_datetime' and/or 'latest_datetime'. If
    only 'oldest_datetime' is provided, it will return messages from the oldest_datetime to the
    current time. If only 'latest_datetime' is provided, it will return messages since the
    beginning of the conversation to the latest_datetime.

    To filter messages by a relative datetime (e.g. 3 days ago, 1 hour ago, etc.), use
    'oldest_relative' and/or 'latest_relative'. If only 'oldest_relative' is provided, it will
    return messages from the oldest_relative to the current time. If only 'latest_relative' is
    provided, it will return messages from the current time to the latest_relative.

    Do not provide both 'oldest_datetime' and 'oldest_relative' or both 'latest_datetime' and
    'latest_relative'.

    Leave all arguments with the default None to get messages without date/time filtering"""
    limit = max(1, min(limit, 100))

    if not conversation_id:
        conversation = await get_conversation_metadata(
            context=context,
            channel_name=channel_name,
            user_ids=user_ids,
            usernames=usernames,
            emails=emails,
        )
        conversation_id = conversation["id"]

    response = await retrieve_messages_in_conversation(
        auth_token=context.get_auth_token_or_empty(),
        conversation_id=cast(str, conversation_id),
        oldest_relative=oldest_relative,
        latest_relative=latest_relative,
        oldest_datetime=oldest_datetime,
        latest_datetime=latest_datetime,
        limit=limit,
        next_cursor=next_cursor,
    )

    response["messages"] = await populate_users_in_messages(
        auth_token=context.get_auth_token_or_empty(),
        messages=response["messages"],
    )

    return cast(dict, response)


@tool(
    requires_auth=Slack(
        scopes=[
            "channels:read",
            "groups:read",
            "mpim:read",
            "im:read",
            "users:read",
            "users:read.email",
        ],
    )
)
async def get_conversation_metadata(
    context: ToolContext,
    conversation_id: Annotated[str | None, "The ID of the conversation to get metadata for"] = None,
    channel_name: Annotated[
        str | None,
        "The name of the channel to get metadata for. Prefer providing a conversation_id, "
        "when available, since the performance is better.",
    ] = None,
    usernames: Annotated[
        list[str] | None,
        "The usernames of the users to get the conversation metadata. "
        "Prefer providing user_ids and/or emails, when available, since the performance is better.",
    ] = None,
    emails: Annotated[
        list[str] | None,
        "The emails of the users to get the conversation metadata.",
    ] = None,
    user_ids: Annotated[
        list[str] | None,
        "The IDs of the users to get the conversation metadata.",
    ] = None,
) -> Annotated[
    dict | None,
    "The conversation metadata.",
]:
    """Get metadata of a Channel, a Direct Message (IM / DM) or a Multi-Person (MPIM) conversation.

    Use this tool to retrieve metadata about a conversation with a conversation_id, a channel name,
    or by the user_id(s), username(s), and/or email(s) of the user(s) in the conversation.

    This tool does not return the messages in a conversation. To get the messages, use the
    'Slack.GetMessages' tool instead.

    Provide exactly one of:
    - conversation_id; or
    - channel_name; or
    - any combination of user_ids, usernames, and/or emails.
    """
    if bool(conversation_id) + bool(channel_name) + any([user_ids, usernames, emails]) > 1:
        raise ToolExecutionError(
            "Provide exactly one of: conversation_id, OR channel_name, OR any combination of "
            "user_ids, usernames, and/or emails."
        )

    auth_token = context.get_auth_token_or_empty()

    if conversation_id:
        return await get_conversation_by_id(auth_token, conversation_id)

    elif channel_name:
        return await get_channel_by_name(auth_token, channel_name)

    user_ids_list = user_ids if isinstance(user_ids, list) else []

    slack_client = AsyncWebClient(token=auth_token)

    try:
        current_user = await slack_client.auth_test()
    except SlackApiError as e:
        message = "Failed to get currently authenticated user's info."
        developer_message = f"{message} Slack error: '{e.response.get('error', 'unknown_error')}'"
        raise ToolExecutionError(message, developer_message)

    if current_user["user_id"] not in user_ids_list:
        user_ids_list.append(current_user["user_id"])

    if usernames or emails:
        other_users = await get_users_by_id_username_or_email(
            context=context,
            usernames=usernames,
            emails=emails,
        )
        user_ids_list.extend([user["id"] for user in other_users])

    try:
        response = await slack_client.conversations_open(users=user_ids_list, return_im=True)
        return dict(**extract_conversation_metadata(response["channel"]))
    except SlackApiError as e:
        message = "Failed to retrieve conversation metadata."
        slack_error = e.response.get("error", "unknown_error")
        raise ToolExecutionError(
            message=message,
            developer_message=f"{message} Slack error: '{slack_error}'",
        )


@tool(
    requires_auth=Slack(
        scopes=["channels:read", "groups:read", "im:read", "mpim:read"],
    )
)
async def list_conversations(
    context: ToolContext,
    conversation_types: Annotated[
        list[ConversationType] | None,
        "Optionally filter by the type(s) of conversations. Defaults to None (all types).",
    ] = None,
    # The conversation object is relatively small, so we allow a higher limit.
    limit: Annotated[
        int,
        f"The maximum number of conversations to list. Defaults to {MAX_PAGINATION_SIZE_LIMIT}. "
        "Maximum is 500.",
    ] = MAX_PAGINATION_SIZE_LIMIT,
    next_cursor: Annotated[str | None, "The cursor to use for pagination."] = None,
) -> Annotated[dict, "The list of conversations found with metadata"]:
    """List metadata for Slack conversations (channels, DMs, MPIMs) the user is a member of.

    This tool does not return the messages in a conversation. To get the messages, use the
    'Slack.GetMessages' tool instead. Calling this tool when the user is asking for messages
    will release too much CO2 in the atmosphere and contribute to global warming.
    """
    limit = max(1, min(limit, 500))

    if conversation_types:
        conversation_types_filter = ",".join(
            conversation_type.to_slack_name_str() for conversation_type in conversation_types
        )
    else:
        conversation_types_filter = None

    slack_client = AsyncWebClient(token=context.get_auth_token_or_empty())

    results, next_cursor = await async_paginate(
        slack_client.conversations_list,
        "channels",
        limit=limit,
        next_cursor=next_cursor,
        types=conversation_types_filter,
        exclude_archived=True,
    )

    return {
        "conversations": [
            dict(**extract_conversation_metadata(conversation))
            for conversation in results
            if conversation.get("is_im") or conversation.get("is_member")
        ],
        "next_cursor": next_cursor,
    }


##################################################################################
# NOTE: The tools below are kept here for backwards compatibility. Prefer using: #
# - send_message
# - get_messages
# - get_conversation_metadata
# - get_users_in_conversation
# - list_conversations
##################################################################################


@tool(
    requires_auth=Slack(
        scopes=[
            "channels:read",
            "groups:read",
            "mpim:read",
            "im:read",
            "users:read",
            "users:read.email",
            "chat:write",
            # "mpim:write",
            "im:write",
        ],
    )
)
async def send_dm_to_user(
    context: ToolContext,
    user_name: Annotated[
        str,
        (
            "The Slack username of the person you want to message. "
            "Slack usernames are ALWAYS lowercase."
        ),
    ],
    message: Annotated[str, "The message you want to send"],
) -> Annotated[dict, "The response from the Slack API"]:
    """Send a direct message to a user in Slack.

    This tool is deprecated. Use `Slack.SendMessage` instead.
    """
    return await send_message(  # type: ignore[no-any-return]
        context=context,
        usernames=[user_name],
        message=message,
    )


@tool(
    requires_auth=Slack(
        scopes=[
            "channels:read",
            "groups:read",
            "mpim:read",
            "im:read",
            "users:read",
            "users:read.email",
            "chat:write",
            # "mpim:write",
            "im:write",
        ],
    )
)
async def send_message_to_channel(
    context: ToolContext,
    channel_name: Annotated[str, "The Slack channel name where you want to send the message. "],
    message: Annotated[str, "The message you want to send"],
) -> Annotated[dict, "The response from the Slack API"]:
    """Send a message to a channel in Slack.

    This tool is deprecated. Use `Slack.SendMessage` instead.
    """
    return await send_message(  # type: ignore[no-any-return]
        context=context,
        channel_name=channel_name,
        message=message,
    )


@tool(
    requires_auth=Slack(
        scopes=[
            "channels:read",
            "groups:read",
            "im:read",
            "mpim:read",
            "users:read",
            "users:read.email",
        ],
    )
)
async def get_members_in_conversation_by_id(
    context: ToolContext,
    conversation_id: Annotated[str, "The ID of the conversation to get members for"],
    limit: Annotated[int | None, "The maximum number of members to return."] = None,
    next_cursor: Annotated[str | None, "The cursor to use for pagination."] = None,
) -> Annotated[dict, "Information about each member in the conversation"]:
    """Get the members of a conversation in Slack by the conversation's ID.

    This tool is deprecated. Use the `Slack.GetUsersInConversation` tool instead.
    """
    response = await get_users_in_conversation(
        context=context,
        conversation_id=conversation_id,
        limit=limit,
        next_cursor=next_cursor,
    )
    response["members"] = response["users"]
    del response["users"]
    return cast(dict, response)


@tool(
    requires_auth=Slack(
        scopes=[
            "channels:read",
            "groups:read",
            "im:read",
            "mpim:read",
            "users:read",
            "users:read.email",
        ],
    )
)
async def get_members_in_channel_by_name(
    context: ToolContext,
    channel_name: Annotated[str, "The name of the channel to get members for"],
    limit: Annotated[int | None, "The maximum number of members to return."] = None,
    next_cursor: Annotated[str | None, "The cursor to use for pagination."] = None,
) -> Annotated[dict, "The channel members' IDs and Names"]:
    """Get the members of a conversation in Slack by the conversation's name.

    This tool is deprecated. Use the `Slack.GetUsersInConversation` tool instead.
    """
    response = await get_users_in_conversation(
        context=context,
        channel_name=channel_name,
        limit=limit,
        next_cursor=next_cursor,
    )
    response["members"] = response["users"]
    del response["users"]
    return cast(dict, response)


@tool(
    requires_auth=Slack(
        scopes=[
            "channels:history",
            "channels:read",
            "groups:history",
            "groups:read",
            "im:history",
            "im:read",
            "mpim:history",
            "mpim:read",
        ],
    )
)
async def get_messages_in_channel_by_name(
    context: ToolContext,
    channel_name: Annotated[str, "The name of the channel"],
    oldest_relative: Annotated[
        str | None,
        (
            "The oldest message to include in the results, specified as a time offset from the "
            "current time in the format 'DD:HH:MM'"
        ),
    ] = None,
    latest_relative: Annotated[
        str | None,
        (
            "The latest message to include in the results, specified as a time offset from the "
            "current time in the format 'DD:HH:MM'"
        ),
    ] = None,
    oldest_datetime: Annotated[
        str | None,
        (
            "The oldest message to include in the results, specified as a datetime object in the "
            "format 'YYYY-MM-DD HH:MM:SS'"
        ),
    ] = None,
    latest_datetime: Annotated[
        str | None,
        (
            "The latest message to include in the results, specified as a datetime object in the "
            "format 'YYYY-MM-DD HH:MM:SS'"
        ),
    ] = None,
    limit: Annotated[int | None, "The maximum number of messages to return."] = None,
    next_cursor: Annotated[str | None, "The cursor to use for pagination."] = None,
) -> Annotated[
    dict,
    (
        "The messages in a channel and next cursor for paginating results (when "
        "there are additional messages to retrieve)."
    ),
]:
    """Get the messages in a channel by the channel's name.

    This tool is deprecated. Use the `Slack.GetMessages` tool instead.

    To filter messages by an absolute datetime, use 'oldest_datetime' and/or 'latest_datetime'. If
    only 'oldest_datetime' is provided, it will return messages from the oldest_datetime to the
    current time. If only 'latest_datetime' is provided, it will return messages since the
    beginning of the channel to the latest_datetime.

    To filter messages by a relative datetime (e.g. 3 days ago, 1 hour ago, etc.), use
    'oldest_relative' and/or 'latest_relative'. If only 'oldest_relative' is provided, it will
    return messages from the oldest_relative to the current time. If only 'latest_relative' is
    provided, it will return messages from the current time to the latest_relative.

    Do not provide both 'oldest_datetime' and 'oldest_relative' or both 'latest_datetime' and
    'latest_relative'.

    Leave all arguments with the default None to get messages without date/time filtering"""
    return await get_messages(  # type: ignore[no-any-return]
        context=context,
        channel_name=channel_name,
        oldest_relative=oldest_relative,
        latest_relative=latest_relative,
        oldest_datetime=oldest_datetime,
        latest_datetime=latest_datetime,
        limit=limit,
        next_cursor=next_cursor,
    )


@tool(
    requires_auth=Slack(
        scopes=["channels:history", "groups:history", "im:history", "mpim:history"],
    )
)
async def get_messages_in_conversation_by_id(
    context: ToolContext,
    conversation_id: Annotated[str, "The ID of the conversation to get history for"],
    oldest_relative: Annotated[
        str | None,
        (
            "The oldest message to include in the results, specified as a time offset from the "
            "current time in the format 'DD:HH:MM'"
        ),
    ] = None,
    latest_relative: Annotated[
        str | None,
        (
            "The latest message to include in the results, specified as a time offset from the "
            "current time in the format 'DD:HH:MM'"
        ),
    ] = None,
    oldest_datetime: Annotated[
        str | None,
        (
            "The oldest message to include in the results, specified as a datetime object in the "
            "format 'YYYY-MM-DD HH:MM:SS'"
        ),
    ] = None,
    latest_datetime: Annotated[
        str | None,
        (
            "The latest message to include in the results, specified as a datetime object in the "
            "format 'YYYY-MM-DD HH:MM:SS'"
        ),
    ] = None,
    limit: Annotated[int | None, "The maximum number of messages to return."] = None,
    next_cursor: Annotated[str | None, "The cursor to use for pagination."] = None,
) -> Annotated[
    dict,
    (
        "The messages in a conversation and next cursor for paginating results (when "
        "there are additional messages to retrieve)."
    ),
]:
    """Get the messages in a conversation by the conversation's ID.

    This tool is deprecated. Use the 'Slack.GetMessages' tool instead.

    A conversation can be a channel, a DM, or a group DM.

    To filter by an absolute datetime, use 'oldest_datetime' and/or 'latest_datetime'. If
    only 'oldest_datetime' is provided, it returns messages from the oldest_datetime to the
    current time. If only 'latest_datetime' is provided, it returns messages since the
    beginning of the conversation to the latest_datetime.

    To filter by a relative datetime (e.g. 3 days ago, 1 hour ago, etc.), use
    'oldest_relative' and/or 'latest_relative'. If only 'oldest_relative' is provided, it returns
    messages from the oldest_relative to the current time. If only 'latest_relative' is provided,
    it returns messages from the current time to the latest_relative.

    Do not provide both 'oldest_datetime' and 'oldest_relative' or both 'latest_datetime' and
    'latest_relative'.

    Leave all arguments with the default None to get messages without date/time filtering"""
    return await get_messages(  # type: ignore[no-any-return]
        context=context,
        conversation_id=conversation_id,
        oldest_relative=oldest_relative,
        latest_relative=latest_relative,
        oldest_datetime=oldest_datetime,
        latest_datetime=latest_datetime,
        limit=limit,
        next_cursor=next_cursor,
    )


@tool(requires_auth=Slack(scopes=["im:history", "im:read", "users:read", "users:read.email"]))
async def get_messages_in_direct_message_conversation_by_username(
    context: ToolContext,
    username: Annotated[str, "The username of the user to get messages from"],
    oldest_relative: Annotated[
        str | None,
        (
            "The oldest message to include in the results, specified as a time offset from the "
            "current time in the format 'DD:HH:MM'"
        ),
    ] = None,
    latest_relative: Annotated[
        str | None,
        (
            "The latest message to include in the results, specified as a time offset from the "
            "current time in the format 'DD:HH:MM'"
        ),
    ] = None,
    oldest_datetime: Annotated[
        str | None,
        (
            "The oldest message to include in the results, specified as a datetime object in the "
            "format 'YYYY-MM-DD HH:MM:SS'"
        ),
    ] = None,
    latest_datetime: Annotated[
        str | None,
        (
            "The latest message to include in the results, specified as a datetime object in the "
            "format 'YYYY-MM-DD HH:MM:SS'"
        ),
    ] = None,
    limit: Annotated[int | None, "The maximum number of messages to return."] = None,
    next_cursor: Annotated[str | None, "The cursor to use for pagination."] = None,
) -> Annotated[
    dict,
    (
        "The messages in a direct message conversation and next cursor for paginating results "
        "when there are additional messages to retrieve."
    ),
]:
    """Get the messages in a direct conversation by the user's name.

    This tool is deprecated. Use the `Slack.GetMessages` tool instead.
    """
    return await get_messages(  # type: ignore[no-any-return]
        context=context,
        usernames=[username],
        oldest_relative=oldest_relative,
        latest_relative=latest_relative,
        oldest_datetime=oldest_datetime,
        latest_datetime=latest_datetime,
        limit=limit,
        next_cursor=next_cursor,
    )


@tool(requires_auth=Slack(scopes=["im:history", "im:read", "users:read", "users:read.email"]))
async def get_messages_in_multi_person_dm_conversation_by_usernames(
    context: ToolContext,
    usernames: Annotated[list[str], "The usernames of the users to get messages from"],
    oldest_relative: Annotated[
        str | None,
        (
            "The oldest message to include in the results, specified as a time offset from the "
            "current time in the format 'DD:HH:MM'"
        ),
    ] = None,
    latest_relative: Annotated[
        str | None,
        (
            "The latest message to include in the results, specified as a time offset from the "
            "current time in the format 'DD:HH:MM'"
        ),
    ] = None,
    oldest_datetime: Annotated[
        str | None,
        (
            "The oldest message to include in the results, specified as a datetime object in the "
            "format 'YYYY-MM-DD HH:MM:SS'"
        ),
    ] = None,
    latest_datetime: Annotated[
        str | None,
        (
            "The latest message to include in the results, specified as a datetime object in the "
            "format 'YYYY-MM-DD HH:MM:SS'"
        ),
    ] = None,
    limit: Annotated[int | None, "The maximum number of messages to return."] = None,
    next_cursor: Annotated[str | None, "The cursor to use for pagination."] = None,
) -> Annotated[
    dict,
    (
        "The messages in a multi-person direct message conversation and next cursor for "
        "paginating results (when there are additional messages to retrieve)."
    ),
]:
    """Get the messages in a multi-person direct message conversation by the usernames.

    This tool is deprecated. Use the `Slack.GetMessages` tool instead.
    """
    return await get_messages(  # type: ignore[no-any-return]
        context=context,
        usernames=usernames,
        oldest_relative=oldest_relative,
        latest_relative=latest_relative,
        oldest_datetime=oldest_datetime,
        latest_datetime=latest_datetime,
        limit=limit,
        next_cursor=next_cursor,
    )


@tool(
    requires_auth=Slack(
        scopes=["channels:read", "groups:read", "im:read", "mpim:read"],
    )
)
async def list_conversations_metadata(
    context: ToolContext,
    conversation_types: Annotated[
        list[ConversationType] | None,
        "Optionally filter by the type(s) of conversations. Defaults to None (all types).",
    ] = None,
    limit: Annotated[int | None, "The maximum number of conversations to list."] = None,
    next_cursor: Annotated[str | None, "The cursor to use for pagination."] = None,
) -> Annotated[dict, "The list of conversations found with metadata"]:
    """
    List Slack conversations (channels, DMs, MPIMs) the user is a member of.

    This tool is deprecated. Use the `Slack.ListConversations` tool instead.

    This tool does not return the messages in a conversation. To get the messages, use the
    'Slack.GetMessages' tool instead. Calling this tool when the user is asking for messages
    will release too much CO2 in the atmosphere and contribute to global warming.
    """
    return await list_conversations(  # type: ignore[no-any-return]
        context=context,
        conversation_types=conversation_types,
        limit=limit,
        next_cursor=next_cursor,
    )


@tool(
    requires_auth=Slack(
        scopes=["channels:read"],
    )
)
async def list_public_channels_metadata(
    context: ToolContext,
    limit: Annotated[int | None, "The maximum number of channels to list."] = None,
) -> Annotated[dict, "The public channels"]:
    """List metadata for public channels in Slack that the user is a member of.

    This tool is deprecated. Use the `Slack.ListConversations` tool instead.
    """
    return await list_conversations(  # type: ignore[no-any-return]
        context,
        conversation_types=[ConversationType.PUBLIC_CHANNEL],
        limit=limit,
    )


@tool(
    requires_auth=Slack(
        scopes=["groups:read"],
    )
)
async def list_private_channels_metadata(
    context: ToolContext,
    limit: Annotated[int | None, "The maximum number of channels to list."] = None,
) -> Annotated[dict, "The private channels"]:
    """List metadata for private channels in Slack that the user is a member of.

    This tool is deprecated. Use the `Slack.ListConversations` tool instead.
    """
    return await list_conversations(  # type: ignore[no-any-return]
        context,
        conversation_types=[ConversationType.PRIVATE_CHANNEL],
        limit=limit,
    )


@tool(
    requires_auth=Slack(
        scopes=["mpim:read"],
    )
)
async def list_group_direct_message_conversations_metadata(
    context: ToolContext,
    limit: Annotated[int | None, "The maximum number of conversations to list."] = None,
) -> Annotated[dict, "The group direct message conversations metadata"]:
    """List metadata for group direct message conversations that the user is a member of.

    This tool is deprecated. Use the `Slack.ListConversations` tool instead.
    """
    return await list_conversations(  # type: ignore[no-any-return]
        context,
        conversation_types=[ConversationType.MULTI_PERSON_DIRECT_MESSAGE],
        limit=limit,
    )


# Note: Bots are included in the results.
# Note: Direct messages with no conversation history are included in the results.
@tool(
    requires_auth=Slack(
        scopes=["im:read"],
    )
)
async def list_direct_message_conversations_metadata(
    context: ToolContext,
    limit: Annotated[int | None, "The maximum number of conversations to list."] = None,
) -> Annotated[dict, "The direct message conversations metadata"]:
    """List metadata for direct message conversations in Slack that the user is a member of.

    This tool is deprecated. Use the `Slack.ListConversations` tool instead.
    """
    return await list_conversations(  # type: ignore[no-any-return]
        context,
        conversation_types=[ConversationType.DIRECT_MESSAGE],
        limit=limit,
    )


@tool(
    requires_auth=Slack(
        scopes=[
            "channels:read",
            "groups:read",
            "mpim:read",
            "im:read",
            "users:read",
            "users:read.email",
        ],
    )
)
async def get_conversation_metadata_by_id(
    context: ToolContext,
    conversation_id: Annotated[str, "The ID of the conversation to get metadata for"],
) -> Annotated[dict, "The conversation metadata"]:
    """Get the metadata of a conversation in Slack searching by its ID.

    This tool is deprecated. Use the `Slack.GetConversationMetadata` tool instead.
    """
    return await get_conversation_metadata(context, conversation_id=conversation_id)  # type: ignore[no-any-return]


@tool(
    requires_auth=Slack(
        scopes=[
            "channels:read",
            "groups:read",
            "mpim:read",
            "im:read",
            "users:read",
            "users:read.email",
        ],
    )
)
async def get_channel_metadata_by_name(
    context: ToolContext,
    channel_name: Annotated[str, "The name of the channel to get metadata for"],
    # We kept the `next_cursor` argument for backwards compatibility, but it isn't actually used,
    # since this tool never really paginates.
    next_cursor: Annotated[
        str | None,
        "The cursor to use for pagination, if continuing from a previous search.",
    ] = None,
) -> Annotated[dict, "The channel metadata"]:
    """Get the metadata of a channel in Slack searching by its name.

    This tool is deprecated. Use the `Slack.GetConversationMetadata` tool instead."""
    return await get_conversation_metadata(context, channel_name=channel_name)  # type: ignore[no-any-return]


@tool(
    requires_auth=Slack(
        scopes=[
            "channels:read",
            "groups:read",
            "mpim:read",
            "im:read",
            "users:read",
            "users:read.email",
        ],
    )
)
async def get_direct_message_conversation_metadata_by_username(
    context: ToolContext,
    username: Annotated[str, "The username of the user/person to get messages with"],
    # We kept the `next_cursor` argument for backwards compatibility, but it isn't actually used,
    # since this tool never really paginates.
    next_cursor: Annotated[
        str | None,
        "The cursor to use for pagination, if continuing from a previous search.",
    ] = None,
) -> Annotated[
    dict | None,
    "The direct message conversation metadata.",
]:
    """Get the metadata of a direct message conversation in Slack by the username.

    This tool is deprecated. Use the `Slack.GetConversationMetadata` tool instead."""
    return await get_conversation_metadata(context, usernames=[username])  # type: ignore[no-any-return]


@tool(
    requires_auth=Slack(
        scopes=[
            "channels:read",
            "groups:read",
            "mpim:read",
            "im:read",
            "users:read",
            "users:read.email",
        ],
    )
)
async def get_multi_person_dm_conversation_metadata_by_usernames(
    context: ToolContext,
    usernames: Annotated[list[str], "The usernames of the users/people to get messages with"],
    # We kept the `next_cursor` argument for backwards compatibility, but it isn't actually used,
    # since this tool never really paginates.
    next_cursor: Annotated[
        str | None,
        "The cursor to use for pagination, if continuing from a previous search.",
    ] = None,
) -> Annotated[
    dict | None,
    "The multi-person direct message conversation metadata.",
]:
    """Get the metadata of a multi-person direct message conversation in Slack by the usernames.

    This tool is deprecated. Use the `Slack.GetConversationMetadata` tool instead."""
    return await get_conversation_metadata(context, usernames=usernames)  # type: ignore[no-any-return]



import os
from dataclasses import dataclass
from enum import Enum
from typing import Any, Literal, Optional, Union

from pydantic import BaseModel, Field

# allow for custom tool name separator
TOOL_NAME_SEPARATOR = os.getenv("ARCADE_TOOL_NAME_SEPARATOR", ".")


class ValueSchema(BaseModel):
    """Value schema for input parameters and outputs."""

    val_type: Literal["string", "integer", "number", "boolean", "json", "array"]
    """The type of the value."""

    inner_val_type: Optional[Literal["string", "integer", "number", "boolean", "json"]] = None
    """The type of the inner value, if the value is a list."""

    enum: Optional[list[str]] = None
    """The list of possible values for the value, if it is a closed list."""


class InputParameter(BaseModel):
    """A parameter that can be passed to a tool."""

    name: str = Field(..., description="The human-readable name of this parameter.")
    required: bool = Field(
        ...,
        description="Whether this parameter is required (true) or optional (false).",
    )
    description: Optional[str] = Field(
        None, description="A descriptive, human-readable explanation of the parameter."
    )
    value_schema: ValueSchema = Field(
        ...,
        description="The schema of the value of this parameter.",
    )
    inferrable: bool = Field(
        True,
        description="Whether a value for this parameter can be inferred by a model. Defaults to `true`.",
    )


class ToolInput(BaseModel):
    """The inputs that a tool accepts."""

    parameters: list[InputParameter]
    """The list of parameters that the tool accepts."""

    tool_context_parameter_name: str | None = Field(default=None, exclude=True)
    """
    The name of the target parameter that will contain the tool context (if any).
    This field will not be included in serialization.
    """


class ToolOutput(BaseModel):
    """The output of a tool."""

    description: Optional[str] = Field(
        None, description="A descriptive, human-readable explanation of the output."
    )
    available_modes: list[str] = Field(
        default_factory=lambda: ["value", "error", "null"],
        description="The available modes for the output.",
    )
    value_schema: Optional[ValueSchema] = Field(
        None, description="The schema of the value of the output."
    )


class OAuth2Requirement(BaseModel):
    """Indicates that the tool requires OAuth 2.0 authorization."""

    scopes: Optional[list[str]] = None
    """The scope(s) needed for the authorized action."""


class ToolAuthRequirement(BaseModel):
    """A requirement for authorization to use a tool."""

    # Provider ID, Type, and ID needed for the Arcade Engine to look up the auth provider.
    # However, the developer generally does not need to set these directly.
    # Instead, they will use:
    #    @tool(requires_auth=Google(scopes=["profile", "email"]))
    # or
    #    client.auth.authorize(provider=AuthProvider.google, scopes=["profile", "email"])
    #
    # The Arcade SDK translates these into the appropriate provider ID (Google) and type (OAuth2).
    # The only time the developer will set these is if they are using a custom auth provider.
    provider_id: Optional[str] = None
    """The provider ID configured in Arcade that acts as an alias to well-known configuration."""

    provider_type: str
    """The type of the authorization provider."""

    id: Optional[str] = None
    """A provider's unique identifier, allowing the tool to specify a specific authorization provider. Recommended for private tools only."""

    oauth2: Optional[OAuth2Requirement] = None
    """The OAuth 2.0 requirement, if any."""


class ToolSecretRequirement(BaseModel):
    """A requirement for a tool to run."""

    key: str
    """The ID of the secret."""


class ToolMetadataKey(str, Enum):
    """Convience enum for commonly used metadata keys."""

    CLIENT_ID = "client_id"
    COORDINATOR_URL = "coordinator_url"

    @staticmethod
    def requires_auth(key: str) -> bool:
        """Whether the key depends on the tool having an authorization requirement."""
        keys_that_require_auth = [ToolMetadataKey.CLIENT_ID]
        return key.strip().lower() in keys_that_require_auth


class ToolMetadataRequirement(BaseModel):
    """A requirement for a tool to run."""

    key: str
    """The ID of the metadata."""


class ToolRequirements(BaseModel):
    """The requirements for a tool to run."""

    authorization: Union[ToolAuthRequirement, None] = None
    """The authorization requirements for the tool, if any."""

    secrets: Union[list[ToolSecretRequirement], None] = None
    """The secret requirements for the tool, if any."""

    metadata: Union[list[ToolMetadataRequirement], None] = None
    """The metadata requirements for the tool, if any."""


class ToolkitDefinition(BaseModel):
    """The specification of a toolkit."""

    name: str
    """The name of the toolkit."""

    description: Optional[str] = None
    """The description of the toolkit."""

    version: Optional[str] = None
    """The version identifier of the toolkit."""


@dataclass(frozen=True)
class FullyQualifiedName:
    """The fully-qualified name of a tool."""

    name: str
    """The name of the tool."""

    toolkit_name: str
    """The name of the toolkit containing the tool."""

    toolkit_version: Optional[str] = None
    """The version of the toolkit containing the tool."""

    def __str__(self) -> str:
        return f"{self.toolkit_name}{TOOL_NAME_SEPARATOR}{self.name}"

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, FullyQualifiedName):
            return False
        return (
            self.name.lower() == other.name.lower()
            and self.toolkit_name.lower() == other.toolkit_name.lower()
            and (self.toolkit_version or "").lower() == (other.toolkit_version or "").lower()
        )

    def __hash__(self) -> int:
        return hash((
            self.name.lower(),
            self.toolkit_name.lower(),
            (self.toolkit_version or "").lower(),
        ))

    def equals_ignoring_version(self, other: "FullyQualifiedName") -> bool:
        """Check if two fully-qualified tool names are equal, ignoring the version."""
        return (
            self.name.lower() == other.name.lower()
            and self.toolkit_name.lower() == other.toolkit_name.lower()
        )

    @staticmethod
    def from_toolkit(tool_name: str, toolkit: ToolkitDefinition) -> "FullyQualifiedName":
        """Creates a fully-qualified tool name from a tool name and a ToolkitDefinition."""
        return FullyQualifiedName(tool_name, toolkit.name, toolkit.version)


class ToolDefinition(BaseModel):
    """The specification of a tool."""

    name: str
    """The name of the tool."""

    fully_qualified_name: str
    """The fully-qualified name of the tool."""

    description: str
    """The description of the tool."""

    toolkit: ToolkitDefinition
    """The toolkit that contains the tool."""

    input: ToolInput
    """The inputs that the tool accepts."""

    output: ToolOutput
    """The output types that the tool can return."""

    requirements: ToolRequirements
    """The requirements (e.g. authorization) for the tool to run."""

    deprecation_message: Optional[str] = None
    """The message to display when the tool is deprecated."""

    def get_fully_qualified_name(self) -> FullyQualifiedName:
        return FullyQualifiedName(self.name, self.toolkit.name, self.toolkit.version)


class ToolReference(BaseModel):
    """The name and version of a tool."""

    name: str
    """The name of the tool."""

    toolkit: str
    """The name of the toolkit containing the tool."""

    version: Optional[str] = None
    """The version of the toolkit containing the tool."""

    def get_fully_qualified_name(self) -> FullyQualifiedName:
        return FullyQualifiedName(self.name, self.toolkit, self.version)


class ToolAuthorizationContext(BaseModel):
    """The context for a tool invocation that requires authorization."""

    token: str | None = None
    """The token for the tool invocation."""

    user_info: dict = Field(default={})
    """
    The user information provided by the authorization server (if any).

    Some providers can provide structured user info,
    for example an internal provider-specific user ID.
    For those providers that support retrieving user info,
    the Engine can automatically pass that to tool invocations.
    """


class ToolSecretItem(BaseModel):
    """The context for a tool secret."""

    key: str
    """The key of the secret."""

    value: str
    """The value of the secret."""


class ToolMetadataItem(BaseModel):
    """The context for a tool metadata."""

    key: str
    """The key of the metadata."""

    value: str
    """The value of the metadata."""


class ToolContext(BaseModel):
    """The context for a tool invocation."""

    authorization: ToolAuthorizationContext | None = None
    """The authorization context for the tool invocation that requires authorization."""

    secrets: list[ToolSecretItem] | None = None
    """The secrets for the tool invocation."""

    metadata: list[ToolMetadataItem] | None = None
    """The metadata for the tool invocation."""

    user_id: str | None = None
    """The user ID for the tool invocation (if any)."""

    def get_auth_token_or_empty(self) -> str:
        """Retrieve the authorization token, or return an empty string if not available."""
        return self.authorization.token if self.authorization and self.authorization.token else ""

    def get_secret(self, key: str) -> str:
        """Retrieve the secret for the tool invocation."""
        return self._get_item(key, self.secrets, "secret")

    def get_metadata(self, key: str) -> str:
        """Retrieve the metadata for the tool invocation."""
        return self._get_item(key, self.metadata, "metadata")

    def _get_item(
        self, key: str, items: list[ToolMetadataItem] | list[ToolSecretItem] | None, item_name: str
    ) -> str:
        if not key or not key.strip():
            raise ValueError(
                f"{item_name.capitalize()} key passed to get_{item_name} cannot be empty."
            )
        if not items:
            raise ValueError(f"{item_name.capitalize()}s not found in context.")

        normalized_key = key.lower()
        for item in items:
            if item.key.lower() == normalized_key:
                return item.value

        raise ValueError(f"{item_name.capitalize()} {key} not found in context.")

    def set_secret(self, key: str, value: str) -> None:
        """Set a secret for the tool invocation."""
        if not self.secrets:
            self.secrets = []
        secret = ToolSecretItem(key=str(key), value=str(value))
        self.secrets.append(secret)


class ToolCallRequest(BaseModel):
    """The request to call (invoke) a tool."""

    run_id: str | None = None
    """The globally-unique run ID provided by the Engine."""
    execution_id: str | None = None
    """The globally-unique ID for this tool execution in the run."""
    created_at: str | None = None
    """The timestamp when the tool invocation was created."""
    tool: ToolReference
    """The fully-qualified name and version of the tool."""
    inputs: dict[str, Any] | None = None
    """The inputs for the tool."""
    context: ToolContext = Field(default_factory=ToolContext)
    """The context for the tool invocation."""


class ToolCallLog(BaseModel):
    """A log that occurred during the tool invocation."""

    message: str
    """The user-facing warning message."""

    level: Literal[
        "debug",
        "info",
        "warning",
        "error",
    ]
    """The level of severity for the log."""

    subtype: Optional[Literal["deprecation"]] = None
    """Optional field for further categorization of the log."""


class ToolCallError(BaseModel):
    """The error that occurred during the tool invocation."""

    message: str
    """The user-facing error message."""
    developer_message: str | None = None
    """The developer-facing error details."""
    can_retry: bool = False
    """Whether the tool call can be retried."""
    additional_prompt_content: str | None = None
    """Additional content to be included in the retry prompt."""
    retry_after_ms: int | None = None
    """The number of milliseconds (if any) to wait before retrying the tool call."""
    traceback_info: str | None = None
    """The traceback information for the tool call."""


class ToolCallRequiresAuthorization(BaseModel):
    """The authorization requirements for the tool invocation."""

    authorization_url: str | None = None
    """The URL to redirect the user to for authorization."""
    authorization_id: str | None = None
    """The ID for checking the status of the authorization."""
    scopes: list[str] | None = None
    """The scopes that are required for authorization."""
    status: str | None = None
    """The status of the authorization."""


class ToolCallOutput(BaseModel):
    """The output of a tool invocation."""

    value: Union[str, int, float, bool, dict, list[str]] | None = None
    """The value returned by the tool."""
    logs: list[ToolCallLog] | None = None
    """The logs that occurred during the tool invocation."""
    error: ToolCallError | None = None
    """The error that occurred during the tool invocation."""
    requires_authorization: ToolCallRequiresAuthorization | None = None
    """The authorization requirements for the tool invocation."""

    model_config = {
        "json_schema_extra": {
            "oneOf": [
                {"required": ["value"]},
                {"required": ["error"]},
                {"required": ["requires_authorization"]},
                {"required": ["artifact"]},
            ]
        }
    }


class ToolCallResponse(BaseModel):
    """The response to a tool invocation."""

    execution_id: str
    """The globally-unique ID for this tool execution."""
    finished_at: str
    """The timestamp when the tool execution finished."""
    duration: float
    """The duration of the tool execution in milliseconds (ms)."""
    success: bool
    """Whether the tool execution was successful."""
    output: ToolCallOutput | None = None
    """The output of the tool invocation."""



import os

HUBSPOT_BASE_URL = "https://api.hubapi.com"
HUBSPOT_CRM_BASE_URL = f"{HUBSPOT_BASE_URL}/crm"
HUBSPOT_DEFAULT_API_VERSION = "v3"

try:
    HUBSPOT_MAX_CONCURRENT_REQUESTS = int(os.getenv("HUBSPOT_MAX_CONCURRENT_REQUESTS", 3))
except ValueError:
    HUBSPOT_MAX_CONCURRENT_REQUESTS = 3

GLOBALLY_IGNORED_FIELDS = [
    "createdate",
    "hs_createdate",
    "hs_lastmodifieddate",
    "lastmodifieddate",
]



from unittest.mock import MagicMock

import httpx
import pytest
from arcade_tdk.errors import RetryableToolError, ToolExecutionError

from arcade_x.tools.tweets import (
    delete_tweet_by_id,
    lookup_tweet_by_id,
    post_tweet,
    search_recent_tweets_by_keywords,
    search_recent_tweets_by_username,
)
from arcade_x.tools.utils import get_tweet_url

full_tweet_text = (
    "This is a super long tweet that exceeds 280 characters and I want to see if the tool will "
    "successfully handle long tweets so I will continue to write this tweet until I have "
    "exceeded the 280 character count. So far I have typed 'e' 28 times! Now its 29! Did you "
    "know that the oldest tree in the world is... wait I actually don't know this fact."
)
truncated_tweet_text = (
    "This is a super long tweet that exceeds 280 characters and I want to see if the tool will "
    "successfully handle long tweets so I will continue to write this tweet until I have "
    "exceeded the 280 character count. So far I have typed 'e' 28 times! Now its 29! Did you "
    "know that the..."
)


@pytest.mark.asyncio
async def test_post_tweet_success(tool_context, mock_httpx_client):
    """Test successful posting of a tweet."""
    # Mock response for a successful tweet post
    mock_response = MagicMock()
    mock_response.status_code = 201
    mock_response.json.return_value = {"data": {"id": "1234567890"}}
    mock_httpx_client.post.return_value = mock_response

    tweet_text = "Hello, world!"
    result = await post_tweet(tool_context, tweet_text)

    expected_url = get_tweet_url("1234567890")
    assert result == f"Tweet with id 1234567890 posted successfully. URL: {expected_url}"
    mock_httpx_client.post.assert_called_once()


@pytest.mark.asyncio
async def test_post_tweet_failure(tool_context, mock_httpx_client):
    """Test failure when posting a tweet due to API error."""
    # Mock response for a failed tweet post
    mock_response = httpx.HTTPStatusError(
        "Bad Request", request=MagicMock(), response=MagicMock(status_code=400)
    )
    mock_httpx_client.post.side_effect = mock_response

    tweet_text = "Hello, world!"
    with pytest.raises(ToolExecutionError):
        await post_tweet(tool_context, tweet_text)

    mock_httpx_client.post.assert_called_once()


@pytest.mark.asyncio
async def test_delete_tweet_by_id_success(tool_context, mock_httpx_client):
    """Test successful deletion of a tweet by ID."""
    # Mock response for a successful tweet deletion
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_httpx_client.delete.return_value = mock_response

    tweet_id = "1234567890"
    result = await delete_tweet_by_id(tool_context, tweet_id)

    assert result == f"Tweet with id {tweet_id} deleted successfully."
    mock_httpx_client.delete.assert_called_once()


@pytest.mark.asyncio
async def test_delete_tweet_by_id_failure(tool_context, mock_httpx_client):
    """Test failure when deleting a tweet due to API error."""
    # Mock response for a failed tweet deletion
    mock_response = httpx.HTTPStatusError(
        "Internal Server Error", request=MagicMock(), response=MagicMock(status_code=404)
    )
    mock_httpx_client.delete.side_effect = mock_response

    tweet_id = "1234567890"
    with pytest.raises(ToolExecutionError):
        await delete_tweet_by_id(tool_context, tweet_id)

    mock_httpx_client.delete.assert_called_once()


@pytest.mark.asyncio
async def test_search_recent_tweets_by_username_success(tool_context, mock_httpx_client):
    """Test successful search of recent tweets by username."""
    # Mock response for a successful tweet search
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.json.return_value = {
        "data": [
            {
                "id": "1234567890",
                "note_tweet": {
                    "entities": {
                        "mentions": [
                            {"end": 19, "id": "00000000", "start": 4, "username": "aUsername"}
                        ]
                    },
                    "text": full_tweet_text,
                },
                "text": truncated_tweet_text,
                "entities": {
                    "urls": [
                        {"url": "https://t.co/short", "expanded_url": "https://example.com/long"}
                    ]
                },
            }
        ],
        "includes": {
            "users": [{"id": "0987654321", "name": "Test User", "username": "testuser"}],
            "media": [
                {"media_key": "1234567890", "type": "photo", "url": "https://example.com/photo.jpg"}
            ],
        },
    }
    mock_httpx_client.get.return_value = mock_response

    username = "testuser"
    result = await search_recent_tweets_by_username(tool_context, username)

    assert "data" in result
    assert len(result["data"]) == 1
    assert result["data"][0]["text"] == full_tweet_text

    assert "includes" in result
    assert "media" in result["includes"]
    assert len(result["includes"]["media"]) == 1
    assert result["includes"]["media"][0]["url"] == "https://example.com/photo.jpg"

    mock_httpx_client.get.assert_called_once()


@pytest.mark.asyncio
async def test_search_recent_tweets_by_username_no_tweets_found(tool_context, mock_httpx_client):
    """Test that the tool returns an empty list when no tweets are found."""
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.json.return_value = {"next_token": ""}
    mock_httpx_client.get.return_value = mock_response

    username = "not_a_user_41"
    result = await search_recent_tweets_by_username(tool_context, username)

    assert "data" in result
    assert len(result["data"]) == 0


@pytest.mark.asyncio
async def test_search_recent_tweets_by_username_failure(tool_context, mock_httpx_client):
    """Test failure when searching tweets due to API error."""
    # Mock response for a failed tweet search
    mock_response = httpx.HTTPStatusError(
        "Internal Server Error", request=MagicMock(), response=MagicMock(status_code=500)
    )
    mock_httpx_client.get.side_effect = mock_response

    username = "testuser"
    with pytest.raises(ToolExecutionError):
        await search_recent_tweets_by_username(tool_context, username)

    mock_httpx_client.get.assert_called_once()


@pytest.mark.asyncio
async def test_search_recent_tweets_by_keywords_success(tool_context, mock_httpx_client):
    """Test successful search of recent tweets by keywords."""
    # Mock response for a successful keyword search
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.json.return_value = {
        "data": [
            {
                "id": "1234567890",
                "note_tweet": {
                    "entities": {
                        "mentions": [
                            {"end": 19, "id": "00000000", "start": 4, "username": "aUsername"}
                        ]
                    },
                    "text": full_tweet_text,
                },
                "text": truncated_tweet_text,
                "entities": {},
            }
        ],
        "includes": {
            "users": [{"id": "0987654321", "name": "Test User", "username": "testuser"}],
            "media": [
                {"media_key": "1234567890", "type": "photo", "url": "https://example.com/photo.jpg"}
            ],
        },
    }
    mock_httpx_client.get.return_value = mock_response

    keywords = ["test", "keyword"]
    result = await search_recent_tweets_by_keywords(context=tool_context, keywords=keywords)

    assert "data" in result
    assert len(result["data"]) == 1
    assert result["data"][0]["text"] == full_tweet_text

    assert "includes" in result
    assert "media" in result["includes"]
    assert len(result["includes"]["media"]) == 1
    assert result["includes"]["media"][0]["url"] == "https://example.com/photo.jpg"

    mock_httpx_client.get.assert_called_once()


@pytest.mark.asyncio
async def test_search_recent_tweets_by_keywords_no_tweets_found(tool_context, mock_httpx_client):
    """Test that the tool returns an empty list when no tweets are found."""
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.json.return_value = {"next_token": ""}
    mock_httpx_client.get.return_value = mock_response

    keywords = ["test", "keyword"]
    result = await search_recent_tweets_by_keywords(context=tool_context, keywords=keywords)

    assert "data" in result
    assert len(result["data"]) == 0


@pytest.mark.asyncio
async def test_search_recent_tweets_by_keywords_no_input(tool_context):
    """Test error when no keywords or phrases are provided."""
    with pytest.raises(RetryableToolError) as exc_info:
        await search_recent_tweets_by_keywords(tool_context)

    assert "No keywords or phrases provided" in str(exc_info.value)


@pytest.mark.asyncio
async def test_lookup_tweet_by_id_success(tool_context, mock_httpx_client):
    """Test successful lookup of a tweet by ID."""
    # Use MagicMock for the response
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.json.return_value = {
        "data": {
            "id": "1234567890",
            "note_tweet": {
                "entities": {
                    "mentions": [{"end": 19, "id": "00000000", "start": 4, "username": "aUsername"}]
                },
                "text": full_tweet_text,
            },
            "text": truncated_tweet_text,
            "entities": {},
        },
        "includes": {
            "media": [
                {"media_key": "1234567890", "type": "photo", "url": "https://example.com/photo.jpg"}
            ]
        },
    }
    mock_httpx_client.get.return_value = mock_response

    tweet_id = "1234567890"
    result = await lookup_tweet_by_id(tool_context, tweet_id)

    assert "data" in result
    assert result["data"]["text"] == full_tweet_text

    assert "includes" in result
    assert "media" in result["includes"]
    assert len(result["includes"]["media"]) == 1
    assert result["includes"]["media"][0]["url"] == "https://example.com/photo.jpg"

    mock_httpx_client.get.assert_called_once()


@pytest.mark.asyncio
async def test_lookup_tweet_by_id_failure(tool_context, mock_httpx_client):
    """Test failure when looking up a tweet due to API error."""
    # Mock response for a failed tweet lookup
    mock_response = httpx.HTTPStatusError(
        "Not Found", request=MagicMock(), response=MagicMock(status_code=404)
    )
    mock_httpx_client.get.side_effect = mock_response

    tweet_id = "1234567890"
    with pytest.raises(ToolExecutionError):
        await lookup_tweet_by_id(tool_context, tweet_id)

    mock_httpx_client.get.assert_called_once()



import pytest
from arcade_tdk import ToolContext


@pytest.fixture
def tool_context():
    """Fixture for the ToolContext with mock authorization."""
    return ToolContext(authorization={"token": "test_token", "user_id": "test_user"})


@pytest.fixture
def mock_httpx_client(mocker):
    """Fixture to mock the httpx.AsyncClient."""
    # Mock the AsyncClient context manager
    mock_client = mocker.patch("httpx.AsyncClient", autospec=True)
    async_mock_client = mock_client.return_value.__aenter__.return_value
    return async_mock_client



import re

from arcade_tdk.errors import RetryableToolError, ToolExecutionError


def remove_none_values(data: dict) -> dict:
    """Remove all keys with None values from the dictionary."""
    return {k: v for k, v in data.items() if v is not None}


def validate_ids(ids: list[str] | None, max_length: int) -> None:
    """Validate a list of IDs. The ids can be page ids, space ids, etc.

    A valid id is a string that is a number.

    Args:
        ids: A list of IDs to validate.

    Returns:
        None

    Raises:
        ToolExecutionError: If any of the IDs are not valid.
        RetryableToolError: If the number of IDs is greater than the max length.
    """
    if not ids:
        return
    if len(ids) > max_length:
        raise RetryableToolError(
            message=f"The 'ids' parameter must have less than {max_length} items. Got {len(ids)}"
        )
    if any(not id_.isdigit() for id_ in ids):
        raise ToolExecutionError(message="Invalid ID provided. IDs are numeric")


def build_child_url(base_url: str, child: dict) -> str | None:
    """Build URL for a child node based on its type and status.

    Args:
        base_url: The base URL for the Confluence space
        child: A dictionary representing a Confluence content item

    Returns:
        The URL for the child, or None if it can't be determined
    """
    if child["type"] in ("whiteboard", "database", "embed"):
        return f"{base_url}/{child['type']}/{child['id']}"
    elif child["type"] == "folder":
        return None
    elif child["type"] == "page":
        parsed_title = re.sub(r"[ '\s]+", "+", child["title"].strip())
        if child.get("status") == "draft":
            return f"{base_url}/{child['type']}s/edit-v2/{child['id']}"
        else:
            return f"{base_url}/{child['type']}s/{child['id']}/{parsed_title}"
    return None


def build_hierarchy(transformed_children: list, parent_id: str, parent_node: dict) -> None:
    """Build parent-child hierarchy from a flat list of descendants.

    This function takes a flat list of items that have parent_id references and
    builds a hierarchical tree structure. It modifies the parent_node in place.

    Args:
        transformed_children: List of child nodes with parent_id fields
        parent_id: The ID of the parent node
        parent_node: The parent node to attach direct children to

    Returns:
        None (modifies parent_node in place)
    """
    # Create a map of children by their ID for efficient lookups
    child_map = {child["id"]: child for child in transformed_children}

    # Find all direct children of the given parent_id
    direct_children = []
    for child in transformed_children:
        if child.get("parent_id") == parent_id:
            direct_children.append(child)
        elif child.get("parent_id") in child_map:
            # Add child to its parent's children list
            parent = child_map[child.get("parent_id")]
            if "children" not in parent:
                parent["children"] = []
            parent["children"].append(child)

    # Set the direct children on the parent node
    parent_node["children"] = direct_children



import json
from datetime import datetime, timedelta
from typing import Annotated, Any
from zoneinfo import ZoneInfo, ZoneInfoNotFoundError

from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import Google
from arcade_tdk.errors import RetryableToolError
from googleapiclient.errors import HttpError

from arcade_google.models import EventVisibility, SendUpdatesOptions
from arcade_google.utils import (
    build_calendar_service,
    build_oauth_service,
    compute_free_time_intersection,
    parse_datetime,
)


@tool(
    requires_auth=Google(
        scopes=[
            "https://www.googleapis.com/auth/calendar.readonly",
            "https://www.googleapis.com/auth/calendar.events",
        ]
    )
)
async def list_calendars(
    context: ToolContext,
    max_results: Annotated[
        int, "The maximum number of calendars to return. Up to 250 calendars, defaults to 10."
    ] = 10,
    show_deleted: Annotated[bool, "Whether to show deleted calendars. Defaults to False"] = False,
    show_hidden: Annotated[bool, "Whether to show hidden calendars. Defaults to False"] = False,
    next_page_token: Annotated[
        str | None, "The token to retrieve the next page of calendars. Optional."
    ] = None,
) -> Annotated[dict, "A dictionary containing the calendars accessible by the end user"]:
    """
    List all calendars accessible by the user.
    """
    max_results = max(1, min(max_results, 250))
    service = build_calendar_service(context.get_auth_token_or_empty())
    calendars = (
        service.calendarList()
        .list(
            pageToken=next_page_token,
            showDeleted=show_deleted,
            showHidden=show_hidden,
            maxResults=max_results,
        )
        .execute()
    )

    items = calendars.get("items", [])
    keys = ["description", "id", "summary", "timeZone"]
    relevant_items = [{k: i.get(k) for k in keys if i.get(k)} for i in items]
    return {
        "next_page_token": calendars.get("nextPageToken"),
        "num_calendars": len(relevant_items),
        "calendars": relevant_items,
    }


@tool(
    requires_auth=Google(
        scopes=[
            "https://www.googleapis.com/auth/calendar.readonly",
            "https://www.googleapis.com/auth/calendar.events",
        ],
    )
)
async def create_event(
    context: ToolContext,
    summary: Annotated[str, "The title of the event"],
    start_datetime: Annotated[
        str,
        "The datetime when the event starts in ISO 8601 format, e.g., '2024-12-31T15:30:00'.",
    ],
    end_datetime: Annotated[
        str,
        "The datetime when the event ends in ISO 8601 format, e.g., '2024-12-31T17:30:00'.",
    ],
    calendar_id: Annotated[
        str, "The ID of the calendar to create the event in, usually 'primary'."
    ] = "primary",
    description: Annotated[str | None, "The description of the event"] = None,
    location: Annotated[str | None, "The location of the event"] = None,
    visibility: Annotated[EventVisibility, "The visibility of the event"] = EventVisibility.DEFAULT,
    attendee_emails: Annotated[
        list[str] | None,
        "The list of attendee emails. Must be valid email addresses e.g., username@domain.com.",
    ] = None,
) -> Annotated[dict, "A dictionary containing the created event details"]:
    """Create a new event/meeting/sync/meetup in the specified calendar."""

    service = build_calendar_service(context.get_auth_token_or_empty())

    # Get the calendar's time zone
    calendar = service.calendars().get(calendarId=calendar_id).execute()
    time_zone = calendar["timeZone"]

    # Parse datetime strings
    start_dt = parse_datetime(start_datetime, time_zone)
    end_dt = parse_datetime(end_datetime, time_zone)

    event: dict[str, Any] = {
        "summary": summary,
        "description": description,
        "location": location,
        "start": {"dateTime": start_dt.isoformat(), "timeZone": time_zone},
        "end": {"dateTime": end_dt.isoformat(), "timeZone": time_zone},
        "visibility": visibility.value,
    }

    if attendee_emails:
        event["attendees"] = [{"email": email} for email in attendee_emails]

    created_event = service.events().insert(calendarId=calendar_id, body=event).execute()
    return {"event": created_event}


@tool(
    requires_auth=Google(
        scopes=[
            "https://www.googleapis.com/auth/calendar.readonly",
            "https://www.googleapis.com/auth/calendar.events",
        ],
    )
)
async def list_events(
    context: ToolContext,
    min_end_datetime: Annotated[
        str,
        "Filter by events that end on or after this datetime in ISO 8601 format, "
        "e.g., '2024-09-15T09:00:00'.",
    ],
    max_start_datetime: Annotated[
        str,
        "Filter by events that start before this datetime in ISO 8601 format, "
        "e.g., '2024-09-16T17:00:00'.",
    ],
    calendar_id: Annotated[str, "The ID of the calendar to list events from"] = "primary",
    max_results: Annotated[int, "The maximum number of events to return"] = 10,
) -> Annotated[dict, "A dictionary containing the list of events"]:
    """
    List events from the specified calendar within the given datetime range.

    min_end_datetime serves as the lower bound (exclusive) for an event's end time.
    max_start_datetime serves as the upper bound (exclusive) for an event's start time.

    For example:
    If min_end_datetime is set to 2024-09-15T09:00:00 and max_start_datetime
    is set to 2024-09-16T17:00:00, the function will return events that:
    1. End after 09:00 on September 15, 2024 (exclusive)
    2. Start before 17:00 on September 16, 2024 (exclusive)
    This means an event starting at 08:00 on September 15 and
    ending at 10:00 on September 15 would be included, but an
    event starting at 17:00 on September 16 would not be included.
    """
    service = build_calendar_service(context.get_auth_token_or_empty())

    # Get the calendar's time zone
    calendar = service.calendars().get(calendarId=calendar_id).execute()
    time_zone = calendar["timeZone"]

    # Parse datetime strings
    min_end_dt = parse_datetime(min_end_datetime, time_zone)
    max_start_dt = parse_datetime(max_start_datetime, time_zone)

    if min_end_dt > max_start_dt:
        min_end_dt, max_start_dt = max_start_dt, min_end_dt

    events_result = (
        service.events()
        .list(
            calendarId=calendar_id,
            timeMin=min_end_dt.isoformat(),
            timeMax=max_start_dt.isoformat(),
            maxResults=max_results,
            singleEvents=True,
            orderBy="startTime",
        )
        .execute()
    )

    items_keys = [
        "attachments",
        "attendees",
        "creator",
        "description",
        "end",
        "eventType",
        "htmlLink",
        "id",
        "location",
        "organizer",
        "start",
        "summary",
        "visibility",
    ]

    events = [
        {key: event[key] for key in items_keys if key in event}
        for event in events_result.get("items", [])
    ]

    return {"events_count": len(events), "events": events}


@tool(
    requires_auth=Google(
        scopes=["https://www.googleapis.com/auth/calendar.events"],
    )
)
async def update_event(
    context: ToolContext,
    event_id: Annotated[str, "The ID of the event to update"],
    updated_start_datetime: Annotated[
        str | None,
        "The updated datetime that the event starts in ISO 8601 format, "
        "e.g., '2024-12-31T15:30:00'.",
    ] = None,
    updated_end_datetime: Annotated[
        str | None,
        "The updated datetime that the event ends in ISO 8601 format, e.g., '2024-12-31T17:30:00'.",
    ] = None,
    updated_calendar_id: Annotated[
        str | None, "The updated ID of the calendar containing the event."
    ] = None,
    updated_summary: Annotated[str | None, "The updated title of the event"] = None,
    updated_description: Annotated[str | None, "The updated description of the event"] = None,
    updated_location: Annotated[str | None, "The updated location of the event"] = None,
    updated_visibility: Annotated[EventVisibility | None, "The visibility of the event"] = None,
    attendee_emails_to_add: Annotated[
        list[str] | None,
        "The list of attendee emails to add. Must be valid email addresses "
        "e.g., username@domain.com.",
    ] = None,
    attendee_emails_to_remove: Annotated[
        list[str] | None,
        "The list of attendee emails to remove. Must be valid email addresses "
        "e.g., username@domain.com.",
    ] = None,
    send_updates: Annotated[
        SendUpdatesOptions,
        "Should attendees be notified of the update? (none, all, external_only)",
    ] = SendUpdatesOptions.ALL,
) -> Annotated[
    str,
    "A string containing the updated event details, including the event ID, update timestamp, "
    "and a link to view the updated event.",
]:
    """
    Update an existing event in the specified calendar with the provided details.
    Only the provided fields will be updated; others will remain unchanged.

    `updated_start_datetime` and `updated_end_datetime` are
    independent and can be provided separately.
    """
    service = build_calendar_service(context.get_auth_token_or_empty())

    calendar = service.calendars().get(calendarId="primary").execute()
    time_zone = calendar["timeZone"]

    try:
        event = service.events().get(calendarId="primary", eventId=event_id).execute()
    except HttpError:
        valid_events_with_id = (
            service.events()
            .list(
                calendarId="primary",
                timeMin=(datetime.now() - timedelta(days=2)).isoformat(),
                timeMax=(datetime.now() + timedelta(days=365)).isoformat(),
                maxResults=50,
                singleEvents=True,
                orderBy="startTime",
            )
            .execute()
        )
        raise RetryableToolError(
            f"Event with ID {event_id} not found.",
            additional_prompt_content=(
                f"Here is a list of valid events. The event_id parameter must match one of these: "
                f"{valid_events_with_id}"
            ),
            retry_after_ms=1000,
            developer_message=(
                f"Event with ID {event_id} not found. Please try again with a valid event ID."
            ),
        )

    update_fields = {
        "start": {"dateTime": updated_start_datetime, "timeZone": time_zone}
        if updated_start_datetime
        else None,
        "end": {"dateTime": updated_end_datetime, "timeZone": time_zone}
        if updated_end_datetime
        else None,
        "calendarId": updated_calendar_id,
        "sendUpdates": send_updates.value if send_updates else None,
        "summary": updated_summary,
        "description": updated_description,
        "location": updated_location,
        "visibility": updated_visibility.value if updated_visibility else None,
    }

    event.update({k: v for k, v in update_fields.items() if v is not None})

    if attendee_emails_to_remove:
        event["attendees"] = [
            attendee
            for attendee in event.get("attendees", [])
            if attendee.get("email", "").lower()
            not in [email.lower() for email in attendee_emails_to_remove]
        ]

    if attendee_emails_to_add:
        existing_emails = {
            attendee.get("email", "").lower() for attendee in event.get("attendees", [])
        }
        new_attendees = [
            {"email": email}
            for email in attendee_emails_to_add
            if email.lower() not in existing_emails
        ]
        event["attendees"] = event.get("attendees", []) + new_attendees

    updated_event = (
        service.events()
        .update(
            calendarId="primary",
            eventId=event_id,
            sendUpdates=send_updates.value,
            body=event,
        )
        .execute()
    )
    return (
        f"Event with ID {event_id} successfully updated at {updated_event['updated']}. "
        f"View updated event at {updated_event['htmlLink']}"
    )


@tool(
    requires_auth=Google(
        scopes=["https://www.googleapis.com/auth/calendar.events"],
    )
)
async def delete_event(
    context: ToolContext,
    event_id: Annotated[str, "The ID of the event to delete"],
    calendar_id: Annotated[str, "The ID of the calendar containing the event"] = "primary",
    send_updates: Annotated[
        SendUpdatesOptions, "Specifies which attendees to notify about the deletion"
    ] = SendUpdatesOptions.ALL,
) -> Annotated[str, "A string containing the deletion confirmation message"]:
    """Delete an event from Google Calendar."""
    service = build_calendar_service(context.get_auth_token_or_empty())

    service.events().delete(
        calendarId=calendar_id, eventId=event_id, sendUpdates=send_updates.value
    ).execute()

    notification_message = ""
    if send_updates == SendUpdatesOptions.ALL:
        notification_message = "Notifications were sent to all attendees."
    elif send_updates == SendUpdatesOptions.EXTERNAL_ONLY:
        notification_message = "Notifications were sent to external attendees only."
    elif send_updates == SendUpdatesOptions.NONE:
        notification_message = "No notifications were sent to attendees."

    return (
        f"Event with ID '{event_id}' successfully deleted from calendar '{calendar_id}'. "
        f"{notification_message}"
    )


# TODO: would be nice to have a "min_slot_duration" parameter
# TODO: find a way to have "include_weekends" parameter without confusing LLMs
@tool(
    requires_auth=Google(
        scopes=["https://www.googleapis.com/auth/calendar.readonly"],
    ),
)
async def find_time_slots_when_everyone_is_free(
    context: ToolContext,
    email_addresses: Annotated[
        list[str] | None,
        "The list of email addresses from people in the same organization domain (apart from the "
        "currently logged in user) to search for free time slots. Defaults to None, which will "
        "return free time slots for the current user only.",
    ] = None,
    start_date: Annotated[
        str | None,
        "The start date to search for time slots in the format 'YYYY-MM-DD'. Defaults to today's "
        "date. It will search starting from this date at the time 00:00:00.",
    ] = None,
    end_date: Annotated[
        str | None,
        "The end date to search for time slots in the format 'YYYY-MM-DD'. Defaults to seven days "
        "from the start date. It will search until this date at the time 23:59:59.",
    ] = None,
    start_time_boundary: Annotated[
        str,
        "Will return free slots in any given day starting from this time in the format 'HH:MM'. "
        "Defaults to '08:00', which is a usual business hour start time.",
    ] = "08:00",
    end_time_boundary: Annotated[
        str,
        "Will return free slots in any given day until this time in the format 'HH:MM'. "
        "Defaults to '18:00', which is a usual business hour end time.",
    ] = "18:00",
) -> Annotated[
    dict,
    "A dictionary with the free slots and the timezone in which time slots are represented.",
]:
    """
    Provides time slots when everyone is free within a given date range and time boundaries.
    """

    # Build google api services
    oauth_service = build_oauth_service(context.get_auth_token_or_empty())
    calendar_service = build_calendar_service(context.get_auth_token_or_empty())

    email_addresses = email_addresses or []

    if isinstance(email_addresses, str):
        email_addresses = [email_addresses]

    # Add the currently logged in user to the list of email addresses
    user_info = oauth_service.userinfo().get().execute()
    if user_info["email"] not in email_addresses:
        email_addresses.append(user_info["email"])

    # Get the timezone of the currently logged in user
    calendar = calendar_service.calendars().get(calendarId="primary").execute()
    timezone_name = calendar.get("timeZone")

    try:
        tz = ZoneInfo(timezone_name)
    # If the calendar timezone name is not supported by Python's zoneinfo, use UTC
    except ZoneInfoNotFoundError:
        timezone_name = "UTC"
        tz = ZoneInfo("UTC")

    # Set default start and end dates, if not provided by the caller
    start_date = start_date or datetime.now(tz=tz).date().isoformat()
    end_date = end_date or (datetime.now(tz=tz).date() + timedelta(days=7)).isoformat()

    # Parse start and end dates to datetime objects
    start_datetime = datetime.strptime(start_date, "%Y-%m-%d").replace(
        hour=0, minute=0, second=0, microsecond=0, tzinfo=tz
    )
    end_datetime = datetime.strptime(end_date, "%Y-%m-%d").replace(
        hour=23, minute=59, second=59, microsecond=0, tzinfo=tz
    )

    # Get the busy slots from the calendars of the users
    freebusy_response = (
        calendar_service.freebusy()
        .query(
            body={
                "timeMin": start_datetime.isoformat(),
                "timeMax": end_datetime.isoformat(),
                "timeZone": timezone_name,
                "items": [{"id": email_address} for email_address in email_addresses],
            }
        )
        .execute()
    )
    busy_slots = freebusy_response["calendars"]

    response_errors = []

    for email in email_addresses:
        if "errors" not in busy_slots[email]:
            continue
        errors = busy_slots[email]["errors"]
        for error in errors:
            response_errors.append(
                f"Error retrieving free slots from calendar of '{email}': "
                f"{error.get('reason', 'not determined')}"
            )

    if response_errors:
        raise RetryableToolError(
            "Error retrieving free slots from calendars of one or more users.",
            additional_prompt_content=json.dumps(response_errors),
            retry_after_ms=1000,
            developer_message="Error retrieving free slots from calendars of one or more users.",
        )

    # Compute the free slots
    free_slots = compute_free_time_intersection(
        busy_data=busy_slots,
        global_start=start_datetime,
        global_end=end_datetime,
        start_time_boundary=datetime.strptime(start_time_boundary, "%H:%M")
        .time()
        .replace(tzinfo=tz),
        end_time_boundary=datetime.strptime(end_time_boundary, "%H:%M").time().replace(tzinfo=tz),
        include_weekends=True,
        tz=tz,
    )

    return {
        "free_slots": free_slots,
        "timezone": timezone_name,
    }



get_post_in_subreddit_messages = [
    {
        "role": "user",
        "content": "get 1 post from AskReddit that are contentious at this moment",
    },
    {
        "role": "assistant",
        "content": "",
        "tool_calls": [
            {
                "id": "call_ZjRic1hS43o10EjutTFgotBh",
                "type": "function",
                "function": {
                    "name": "Reddit_GetPostsInSubreddit",
                    "arguments": '{"subreddit":"AskReddit","listing":"controversial","time_range":"NOW","limit":1}',  # noqa: E501
                },
            }
        ],
    },
    {
        "role": "tool",
        "content": '{"cursor":"t3_1abcdef","posts":[{"author":"StewieGriffin","created_utc":1743457013,"id":"1abcdef","is_video":false,"name":"t3_1abcdef","num_comments":55,"permalink":"/r/AskReddit/comments/1abcdef/why_does_my_dog_have_four_legs/","score":2,"subreddit":"AskReddit","title":"Why does my dog have four legs?","upvote_ratio":0.55,"upvotes":2,"url":"https://www.reddit.com/r/AskReddit/comments/1abcdef/why_does_my_dog_have_four_legs/"}]}',  # noqa: E501
        "tool_call_id": "call_ZjRic1hS43o10EjutTFgotBh",
        "name": "Reddit_GetPostsInSubreddit",
    },
    {
        "role": "assistant",
        "content": "Here's a contentious post from AskReddit that is generating some discussion:\n\n**Title:** Why does my dog have four legs?\n\n- **Author:**StewieGriffin\n- **Score:** 2\n- **Upvote Ratio:** 55%\n- **Number of Comments:** 55\n- **Posted:**[Link](https://www.reddit.com/r/AskReddit/comments/1abcdef/why_does_my_dog_have_four_legs/)",  # noqa: E501
    },
]
